import tempfile
import sacc
import sacc.data_types
import sacc.tracers
import numpy as np
import pytest
import os
import pathlib
import urllib
import time
import qp

test_dir = pathlib.Path(__file__).resolve().parent
test_data_dir = test_dir / 'data'


# idea based on TreeCorr tests
def get_from_wiki(url):
    file_name = url.split('/')[-1]
    local_file_name = test_data_dir / file_name
    if not local_file_name.exists():
        print(f"Downlading {url} to data dir")
        try:
            urllib.request.urlretrieve(url, local_file_name)
        except urllib.request.HTTPError as err:
            if err.code == 429:
                print("Rate limit download - waiting 10 seconds to try again")
                time.sleep(10)
                urllib.request.urlretrieve(url, local_file_name)
            else:
                raise
    return local_file_name


def get_filled_sacc():
    s = sacc.Sacc()

    # Tracer
    z = np.arange(0., 1.0, 0.01)
    nz = (z-0.5)**2/0.1**2
    s.add_tracer('NZ', 'source_0', z, nz)
    s.add_tracer('NZ', 'source_1', z, nz,
                 quantity='cluster_density')
    s.add_tracer('NZ', 'source_2', z, nz,
                 quantity='cluster_density')

    for i in range(20):
        ee = 0.1 * i
        tracers = ('source_0',)
        s.add_data_point(sacc.standard_types.count,
                         tracers, ee, ell=10.0*i)
    for i in range(20):
        bb = 0.2 * i
        tracers = ('source_0', 'source_1')
        s.add_data_point(sacc.standard_types.galaxy_shear_cl_bb,
                         tracers, bb, ell=10.0*i)
    for i in range(20):
        ee = 0.3 * i
        tracers = ('source_1', 'source_1', 'source_0')
        s.add_data_point(sacc.standard_types.galaxy_shear_cl_ee,
                         tracers, ee, ell=10.0*i)
    for i in range(20):
        bb = 0.4 * i
        tracers = ('source_2', 'source_2')
        s.add_data_point(sacc.standard_types.galaxy_shear_cl_bb,
                         tracers, bb, ell=10.0*i)

    return s


def test_add_covariance():
    s = get_filled_sacc()
    cov = np.ones((s.mean.size, s.mean.size))
    s.add_covariance(cov)

    with pytest.raises(RuntimeError):
        s.add_covariance(cov)

    s.add_covariance(0 * cov, overwrite=True)
    assert np.all(s.covariance.covmat == 0 * cov)


def test_quantity_warning():
    s = sacc.Sacc()
    with pytest.warns(UserWarning):
        s.add_tracer('Misc', 'source_0',
                     quantity='dummy')


def test_data_type_warning():
    s = sacc.Sacc()
    s.add_tracer('Misc', 'source_0')
    with pytest.warns(UserWarning):
        s.add_data_point('cl_wrong', ('source_0', 'source_0'),
                         0.1, ell=10.)


def test_get_data_types():
    s = get_filled_sacc()
    dt1 = [sacc.standard_types.count, sacc.standard_types.galaxy_shear_cl_bb,
           sacc.standard_types.galaxy_shear_cl_ee]
    dt2 = s.get_data_types()
    assert sorted(dt1) == sorted(dt2)

    dt2 = s.get_data_types(tracers=('source_0', 'source_1'))
    assert [sacc.standard_types.galaxy_shear_cl_bb] == dt2


def test_table_creation_0_instances():
    """If there are no tracer test_table_creation_0_instances, we should
    generate no tables."""
    tables = sacc.tracers.BaseTracer.to_tables([])
    assert len(tables) == 0


def test_table_creation_misc_1_instance():
    t = sacc.tracers.MiscTracer("source_0")
    tables = sacc.tracers.BaseTracer.to_tables([t])
    assert len(tables) == 1


def test_table_creation_misc_2_instances():
    t1 = sacc.tracers.MiscTracer("source_0")
    t2 = sacc.tracers.MiscTracer("source_1")
    tables = sacc.tracers.BaseTracer.to_tables([t1, t2])
    assert len(tables) == 1


def test_table_creation_nz_1_instance():
    t = sacc.tracers.NZTracer("source_0", np.zeros(1), np.zeros(1))
    tables = sacc.tracers.BaseTracer.to_tables([t])
    assert len(tables) == 1


def test_table_creation_nz_2_instances():
    t1 = sacc.tracers.NZTracer("source_0", np.zeros(1), np.zeros(1))
    t2 = sacc.tracers.NZTracer("source_1", np.zeros(1), np.zeros(1))
    tables = sacc.tracers.BaseTracer.to_tables([t1, t2])
    assert len(tables) == 2


def test_table_creation_map_1_instance():
    t = sacc.tracers.MapTracer("source_0", 1, [1.5], [10.0])
    tables = sacc.tracers.BaseTracer.to_tables([t])
    assert len(tables) == 1


def test_table_creation_map_2_instances():
    t1 = sacc.tracers.MapTracer("source_0", 1, [1.5], [10.0])
    t2 = sacc.tracers.MapTracer("source_1", 2, [2.5], [15.0])
    tables = sacc.tracers.BaseTracer.to_tables([t1, t2])
    assert len(tables) == 2


def test_table_creation_numap_1_instance():
    t = sacc.tracers.NuMapTracer("source_0", 1, [1.5], [10.0], np.zeros(1), np.zeros(1))
    tables = sacc.tracers.BaseTracer.to_tables([t])
    assert len(tables) == 2


def test_table_creation_numap_2_instances():
    t1 = sacc.tracers.NuMapTracer("source_0", 1, [1.5], [10.0], np.zeros(1), np.zeros(1))
    t2 = sacc.tracers.NuMapTracer("source_1", 2, [2.5], [15.0], np.zeros(1), np.zeros(1))
    tables = sacc.tracers.BaseTracer.to_tables([t1, t2])
    assert len(tables) == 4


def test_construct():
    s = sacc.Sacc()

    # Tracer
    z = np.arange(0., 1.0, 0.01)
    nz = (z-0.5)**2/0.1**2
    s.add_tracer('NZ', 'source_0', z, nz,
                 quantity='galaxy_density')

    for i in range(20):
        ee = 0.1 * i
        tracers = ('source_0', 'source_0')
        s.add_data_point(sacc.standard_types.cl_00, tracers, ee, ell=10.0*i)


def test_tracers_later():
    s = sacc.Sacc()

    with pytest.raises(ValueError):
        tracers = ('source_0', 'source_0')
        s.add_data_point(sacc.standard_types.galaxy_shear_cl_ee,
                         tracers, 0.0, ell=1)

    s = sacc.Sacc()

    tracers = ('source_0', 'source_0')
    s.add_data_point(sacc.standard_types.galaxy_shear_cl_ee,
                     tracers, 0.0, tracers_later=True, ell=1)


def test_full_cov():
    covmat = np.random.uniform(size=(100, 100))
    C = sacc.covariance.BaseCovariance.make(covmat)
    assert C.size == 100
    assert isinstance(C, sacc.covariance.FullCovariance)
    assert np.all(C.covmat == covmat)
    hdu = C.to_hdu()
    C2 = sacc.covariance.BaseCovariance.from_hdu(hdu)
    assert np.allclose(C.covmat, C2.covmat)


def test_block_cov():
    covmat = [np.random.uniform(size=(50, 50)),
              np.random.uniform(size=(100, 100)),
              np.random.uniform(size=(150, 150))]
    C = sacc.covariance.BaseCovariance.make(covmat)
    assert C.size == 300
    assert isinstance(C, sacc.covariance.BlockDiagonalCovariance)
    hdu = C.to_hdu()
    C2 = sacc.covariance.BaseCovariance.from_hdu(hdu)
    assert len(C2.blocks) == 3
    assert C.block_sizes == C2.block_sizes
    for i in range(3):
        assert np.allclose(C.blocks[i], C2.blocks[i])


def test_misc_tracer():
    md1 = {'potato': 'if_necessary', 'answer': 42, 'height': 1.83}
    md2 = {'potato': 'never'}
    T1 = sacc.BaseTracer.make('Misc', 'tracer1',
                              quantity='generic', metadata=md1)
    T2 = sacc.BaseTracer.make('Misc', 'tracer2',
                              quantity='generic', metadata=md2)
    assert T1.metadata == md1
    assert T2.metadata == md2

    tables = sacc.BaseTracer.to_tables([T1, T2])
    D = sacc.BaseTracer.from_tables(tables)

    T1a = D['tracer1']
    T2a = D['tracer2']
    assert T1a.metadata == md1
    assert T2a.metadata == md2


def test_numap_tracer():
    md1 = {'mac': 'yes', 'cheese': 'of_course', 'quantity': 3}
    md2 = {'mac': 'no', 'cheese': 'no'}
    ell = np.linspace(2, 1000, 1000)
    beam = np.exp(-0.1 * ell * (ell + 1))
    beam_extra = {'err1': np.sin(ell * 0.1)}
    nu = np.linspace(30., 60., 100)
    bandpass = np.ones(100)
    bandpass_extra = {'err1': bandpass * 0.1,
                      'err2': bandpass * 0.05}

    T1 = sacc.BaseTracer.make('NuMap', 'band1', 0,
                              nu, bandpass, ell, beam,
                              quantity='cmb_temperature',
                              bandpass_extra=bandpass_extra,
                              beam_extra=beam_extra,
                              metadata=md2)
    T2 = sacc.BaseTracer.make('NuMap', 'band2', 0,
                              nu, bandpass, ell, beam,
                              quantity='cmb_convergence',
                              bandpass_extra=bandpass_extra,
                              beam_extra=beam_extra,
                              metadata=md1)

    assert T2.metadata == md1
    assert T1.metadata == md2

    tables = sacc.BaseTracer.to_tables([T1, T2])
    D = sacc.BaseTracer.from_tables(tables)

    T1a = D['band1']
    T2a = D['band2']
    assert T1a.metadata == md2
    assert T2a.metadata == md1
    assert np.all(T1a.bandpass_extra['err1'] == 0.1 * bandpass)


def test_map_tracer():
    md1 = {'mac': 'yes', 'cheese': 'of_course', 'quantity': 3}
    md2 = {'mac': 'no', 'cheese': 'no'}
    ell = np.linspace(2, 1000, 1000)
    beam = np.exp(-0.1 * ell * (ell + 1))
    err = np.sin(ell * 0.1)
    beam_extra = {'err1': err}

    T1 = sacc.BaseTracer.make('Map', 'y_milca',
                              0, ell, beam,
                              quantity='cmb_tSZ',
                              beam_extra=beam_extra,
                              metadata=md1)
    T2 = sacc.BaseTracer.make('Map', 'y_nilc',
                              0, ell, beam,
                              quantity='cmb_kSZ',
                              beam_extra=beam_extra,
                              metadata=md2)

    assert T1.metadata == md1
    assert T2.metadata == md2

    tables = sacc.BaseTracer.to_tables([T1, T2])
    D = sacc.BaseTracer.from_tables(tables)

    T1a = D['y_milca']
    T2a = D['y_nilc']
    assert T1a.metadata == md1
    assert T2a.metadata == md2
    assert np.all(T1a.beam_extra['err1'] == err)


def test_nz_tracer():
    md1 = {'potato': 'if_necessary', 'answer': 42, 'height': 1.83}
    md2 = {'potato': 'never'}
    z = np.arange(0., 1., 0.01)
    Nz1 = 1*z  # not a sensible N(z)!
    Nz2 = 2*z  # not a sensible N(z)!

    Nz3 = 3*z
    Nz4 = 4*z

    more_nz = {'v1': Nz3, 'v2': Nz4}

    T1 = sacc.BaseTracer.make('NZ', 'tracer1', z, Nz1,
                              quantity='galaxy_density',
                              extra_columns=more_nz,
                              spin=0,
                              metadata=md1)
    T2 = sacc.BaseTracer.make('NZ', 'tracer2', z, Nz2,
                              quantity='galaxy_shear',
                              spin=2,
                              metadata=md2)
    assert T1.metadata == md1
    assert T2.metadata == md2

    # check that we can make a tracer using lower case
    T3 = sacc.BaseTracer.make('nz', 'tracer2', z, Nz2,
                              quantity='galaxy_shear',
                              spin=2,
                              metadata=md2)
    assert T3.metadata == md2

    tables = sacc.BaseTracer.to_tables([T1, T2])
    D = sacc.BaseTracer.from_tables(tables)

    T1a = D['tracer1']
    T2a = D['tracer2']
    assert T1a.metadata == md1
    assert T2a.metadata == md2

    assert np.all(T1a.extra_columns['v1'] == Nz3)
    assert np.all(T1a.extra_columns['v2'] == Nz4)


def test_mixed_tracers():
    md1 = {'potato': 'never'}
    md2 = {'rank': 'duke'}
    md3 = {'rank': 'earl', 'robes': 78}
    z = np.arange(0., 1., 0.01)
    Nz1 = 1*z  # not a sensible N(z)!
    Nz2 = 2*z
    T1 = sacc.BaseTracer.make('NZ', 'tracer1', z, Nz1,
                              quantity='galaxy_convergence')
    T2 = sacc.BaseTracer.make('NZ', 'tracer2', z, Nz2,
                              quantity='galaxy_shear', metadata=md1)

    M1 = sacc.BaseTracer.make("Misc", "sample1", metadata=md2)
    M2 = sacc.BaseTracer.make("mISC", "sample2", metadata=md3)

    tables = sacc.BaseTracer.to_tables([T1, M1, T2, M2])
    recovered = sacc.BaseTracer.from_tables(tables)
    assert recovered['sample1'].metadata['rank'] == 'duke'
    assert recovered['sample2'].metadata['robes'] == 78
    assert np.all(recovered['tracer1'].nz == Nz1)
    assert recovered['tracer2'].metadata['potato'] == 'never'


def test_inverses():
    N = 25
    C = np.random.uniform(0, 1, size=(N, N))
    C = (C+C.T) + np.eye(N)*20
    M1 = sacc.BaseCovariance.make(C)
    assert M1.size == N
    invC = M1.inverse
    ii = np.dot(invC, C)
    assert np.allclose(ii, np.eye(N))

    blocks = [np.random.uniform(0, 1, size=(5, 5))
              for i in range(5)]
    for b in blocks:
        b += b.T + np.eye(5)*20

    M2 = sacc.BaseCovariance.make(blocks)
    assert M2.size == N
    M2dense = np.zeros((N, N))
    for i in range(5):
        M2dense[i*5:i*5+5, i*5:i*5+5] = blocks[i]
    invC2 = M2.inverse
    ii = np.dot(invC2, M2dense)
    assert np.allclose(ii, np.eye(N))

    d = abs(np.random.uniform(0, 1, size=N))+1
    M3 = sacc.BaseCovariance.make(d)
    assert M3.size == N
    invC3 = M3.inverse
    assert np.count_nonzero(invC3 - np.diag(np.diagonal(invC3))) == 0
    assert np.allclose(invC3.diagonal() * d, 1)


def test_data_point():
    from sacc.data_types import DataPoint
    dt = sacc.data_types.standard_types.galaxy_shearDensity_cl_e
    value = 13.4
    tracers = ('aaa', 'bbb')
    tags = {'ell': 12, 'theta': 14.3}
    d = DataPoint(dt, tracers, value, **tags)
    s = repr(d)
    d2 = eval(s)
    assert d.tracers == d2.tracers
    assert d.tags == d2.tags
    assert d.data_type == d2.data_type
    assert d.value == d2.value


def test_keep_remove():
    s = sacc.Sacc()

    # Tracer
    z = np.arange(0., 1.0, 0.01)
    nz = (z-0.5)**2/0.1**2
    s.add_tracer('NZ', 'source_0', z, nz)
    s.add_tracer('nZ', 'source_1', z, nz,
                 quantity='galaxy_shear', spin=2)
    s.add_tracer('nz', 'source_2', z, nz,
                 quantity='cluster_density')

    for i in range(20):
        ee = 0.1 * i
        tracers = ('source_0', 'source_0')
        s.add_data_point(sacc.standard_types.galaxy_shear_cl_ee,
                         tracers, ee, ell=10.0*i)
    for i in range(20):
        bb = 0.1 * i
        tracers = ('source_1', 'source_1')
        s.add_data_point(sacc.standard_types.galaxy_shear_cl_bb,
                         tracers, bb, ell=10.0*i)
    for i in range(20):
        ee = 0.1 * i
        tracers = ('source_2', 'source_2')
        s.add_data_point(sacc.standard_types.galaxy_shear_cl_ee,
                         tracers, ee, ell=10.0*i)

    # Select by data type
    s2 = s.copy()
    s2.keep_selection(data_type=sacc.standard_types.galaxy_shear_cl_bb)
    assert all(d.data_type == sacc.standard_types.galaxy_shear_cl_bb
               for d in s2.data)
    assert len(s2) == 20

    # From multiple tracers
    s2 = s.copy()
    s2.keep_selection(data_type=sacc.standard_types.galaxy_shear_cl_ee)
    assert all(d.data_type == sacc.standard_types.galaxy_shear_cl_ee
               for d in s2.data)
    assert len(s2) == 40

    # Test removing a single tracer
    s2 = s.copy()
    s2.remove_selection(tracers=('source_1', 'source_1'))
    for i, d in enumerate(s2.data):
        if i < 20:
            assert d.tracers == ('source_0', 'source_0')
        else:
            assert d.tracers == ('source_2', 'source_2')
    assert all(d.data_type == sacc.standard_types.galaxy_shear_cl_ee
               for d in s2.data)
    assert len(s2) == 40

    # Test selecting by tag
    s2 = s.copy()
    s2.remove_selection(ell__lt=55)
    ell = s2.get_tag('ell')
    for e in ell:
        assert e > 55
    s2 = s.copy()
    s2.keep_selection(ell__lt=55)
    ell = s2.get_tag('ell')
    for e in ell:
        assert e < 55

    # Cutting just by index
    s2 = s.copy()
    ind = s2.indices(tracers=('source_1', 'source_1'))
    assert (ind == np.arange(20, 40)).all()

    # multiple selections
    s2 = s.copy()
    ind = s2.indices(tracers=('source_2', 'source_2'), ell__lt=45)
    assert len(ind) == 5


def test_remove_keep_tracers():
    s = get_filled_sacc()

    s.remove_tracers(['source_0'])

    assert ['source_1', 'source_2'] == list(s.tracers.keys())
    assert [('source_2', 'source_2')] == s.get_tracer_combinations()

    s = get_filled_sacc()
    s.keep_tracers(['source_0'])
    assert ['source_0'] == list(s.tracers.keys())
    assert [('source_0',)] == s.get_tracer_combinations()


def test_cutting_block_cov():
    covmat = [np.random.uniform(size=(50, 50)),
              np.random.uniform(size=(100, 100)),
              np.random.uniform(size=(150, 150))]
    C = sacc.covariance.BaseCovariance.make(covmat)
    ind = list(range(50))
    C2 = C.keeping_indices(np.arange(50))
    assert C2.size == len(ind)
    assert np.allclose(C2.get_block(ind), covmat[0])


def test_cutting_block_cov2():
    covmat = [np.random.uniform(size=(50, 50)),
              np.random.uniform(size=(100, 100)),
              np.random.uniform(size=(150, 150))]
    C = sacc.covariance.BaseCovariance.make(covmat)
    ind = list(range(50, 150))
    C2 = C.keeping_indices(np.arange(50, 150))
    assert C2.size == len(ind)
    assert np.allclose(C2.get_block(range(100)), covmat[1])


def test_cutting_full_cov():
    covmat = np.random.uniform(size=(100, 100))
    C = sacc.covariance.BaseCovariance.make(covmat)
    ind = np.arange(10, dtype=int)
    C2 = C.keeping_indices(ind)
    assert np.allclose(C2.get_block(ind),
                       covmat[:10, :10])


def test_cutting_diag_cov():
    diag = np.random.uniform(size=(100,))
    C = sacc.covariance.BaseCovariance.make(diag)
    ind = np.arange(20, dtype=int)
    C2 = C.keeping_indices(ind)
    assert np.allclose(C2.get_block(ind).diagonal(), diag[:20])


def test_parse_data_names():
    for name in sacc.data_types.required_tags_verbose:
        sources, props, stat, sub = sacc.parse_data_type_name(name)
        name2 = sacc.build_data_type_name(sources, props, stat, sub)
        assert name == name2


def test_bandpower_window():
    nb = 20
    nl = 200
    dl = nl // nb
    ells = np.arange(nl)
    w = np.zeros([nb, nl])
    for i in range(nb):
        w[i, i*dl: (i+1)*dl] = 1./dl

    W1 = [sacc.BandpowerWindow(ells, w.T)]

    tables = sacc.BandpowerWindow.to_tables(W1)
    W2 = sacc.BandpowerWindow.from_tables(tables)
    for w1 in W1:
        w2 = W2[id(w1)]
        assert np.all(w1.values == w2.values)
        assert np.all(w1.weight.flatten() == w2.weight.flatten())


def test_tophat_window():
    edges = np.arange(10) * 10
    W1 = [sacc.TopHatWindow(edges[:-1], edges[1:])]

    tables = sacc.TopHatWindow.to_tables(W1)
    W2 = sacc.TopHatWindow.from_tables(tables)
    for w1 in W1:
        w2 = W2[id(w1)]
        assert np.all(w1.min == w2.min)
        assert np.all(w1.max == w2.max)


def test_log_window():
    edges = (np.arange(10) + 1) * 10
    W1 = [sacc.LogTopHatWindow(edges[:-1], edges[1:])]

    tables = sacc.LogTopHatWindow.to_tables(W1)
    W2 = sacc.LogTopHatWindow.from_tables(tables)
    for w1 in W1:
        w2 = W2[id(w1)]
        assert np.all(w1.min == w2.min)
        assert np.all(w1.max == w2.max)


def test_concatenate_covariance():
    v1 = np.array([1., 2., 3.])
    v2 = np.array([4.])
    A = sacc.BaseCovariance.make(v1)
    B = sacc.BaseCovariance.make(v2)
    C = sacc.covariance.concatenate_covariances(A, B)
    assert isinstance(C, sacc.covariance.DiagonalCovariance)
    assert np.allclose(C.diag, [1, 2, 3, 4])

    v1 = np.array([2.])
    v2 = np.array([[3., 0.1], [0.1, 3]])

    A = sacc.BaseCovariance.make(v1)
    B = sacc.BaseCovariance.make(v2)
    C = sacc.covariance.concatenate_covariances(A, B)
    assert isinstance(C, sacc.covariance.BlockDiagonalCovariance)
    test_C = np.array([
        [2.0, 0.0, 0.0],
        [0.0, 3.0, 0.1],
        [0.0, 0.1, 3.0]]
        )
    assert np.allclose(C.dense, test_C)

    v1 = np.array([[2.0, 0.2, ],
                   [0.2, 3.0, ]])
    v2 = np.array([[4.0, -0.2, ],
                   [-0.2, 5.0, ]])
    test_C = np.array([[2.0, 0.2, 0.0, 0.0],
                       [0.2, 3.0, 0.0, 0.0],
                       [0.0, 0.0, 4.0, -0.2],
                       [0.0, 0.0, -0.2, 5.0]])

    A = sacc.BaseCovariance.make(v1)
    B = sacc.BaseCovariance.make(v2)
    C = sacc.covariance.concatenate_covariances(A, B)
    assert isinstance(C, sacc.covariance.BlockDiagonalCovariance)
    assert np.allclose(C.dense, test_C)


def test_concatenate_data():
    s1 = sacc.Sacc()

    # Tracer
    z = np.arange(0., 1.0, 0.01)
    nz = (z-0.5)**2/0.1**2
    s1.add_tracer('NZ', 'source_0', z, nz)

    for i in range(20):
        ee = 0.1 * i
        tracers = ('source_0', 'source_0')
        s1.add_data_point(sacc.standard_types.galaxy_shear_cl_ee,
                          tracers, ee, ell=10.0*i)

    s2 = sacc.Sacc()

    # Tracer
    z = np.arange(0., 1.0, 0.01)
    nz = (z-0.5)**2/0.1**2
    s2.add_tracer('NZ', 'source_0', z, nz,
                  quantity='galaxy_shear', spin=2)

    for i in range(20):
        ee = 0.1 * i
        tracers = ('source_0', 'source_0')
        s2.add_data_point(sacc.standard_types.galaxy_shear_cl_ee,
                          tracers, ee, ell=10.0*i, label='xxx')

    # same tracer
    s3 = sacc.concatenate_data_sets(s1, s2, same_tracers=['source_0'])
    assert ['source_0'] == list(s3.tracers.keys())

    # check data points in right order
    for i in range(20):
        assert s3.data[i].get_tag('ell') == 10.0*i
        assert s3.data[i+20].get_tag('ell') == 10.0*i
        assert s3.data[i].get_tag('label') is None
        assert s3.data[i+20].get_tag('label') == 'xxx'
        t1 = s3.data[i].tracers[0]
        t2 = s3.data[i+20].tracers[0]
        assert t1 == 'source_0'
        assert t1 == t2
        # To make sure the first 'source_0' tracer is used and not rewritten
        s3.get_tracer(t1).quantity == 'generic'

    # name clash
    with pytest.raises(ValueError):
        sacc.concatenate_data_sets(s1, s2)

    s3 = sacc.concatenate_data_sets(s1, s2, labels=['1', '2'])
    assert 'source_0_1' in s3.tracers
    assert 'source_0_2' in s3.tracers
    assert len(s3) == len(s1) + len(s2)

    # check data points in right order
    for i in range(20):
        assert s3.data[i].get_tag('ell') == 10.0*i
        assert s3.data[i+20].get_tag('ell') == 10.0*i
        assert s3.data[i].get_tag('label') == '1'
        assert s3.data[i+20].get_tag('label') == 'xxx_2'
        t1 = s3.data[i].tracers[0]
        t2 = s3.data[i+20].tracers[0]
        assert t1 == 'source_0_1'
        assert t2 == 'source_0_2'
        s3.get_tracer(t1)
        s3.get_tracer(t2)

    # labels + same_tracers
    s4 = sacc.concatenate_data_sets(s3, s3, labels=['x', 'y'],
                                    same_tracers=['source_0_1'])
    trs = ['source_0_1', 'source_0_2_x', 'source_0_2_y']
    assert trs == list(s4.tracers.keys())
    assert s4.mean.size == 2 * s3.mean.size


def test_io():
    s = sacc.Sacc()

    # Tracer
    z = np.arange(0., 1.0, 0.01)
    nz = (z-0.5)**2/0.1**2
    s.add_tracer('NZ', 'source_0', z, nz)

    for i in range(20):
        ee = 0.1 * i
        tracers = ('source_0', 'source_0')
        s.add_data_point(sacc.standard_types.galaxy_shear_cl_ee,
                         tracers, ee, ell=10.0*i)

    with tempfile.TemporaryDirectory() as tmpdir:
        filename = os.path.join(tmpdir, 'test.sacc')
        s.save_fits(filename)
        s2 = sacc.Sacc.load_fits(filename)

    assert len(s2) == 20
    mu = s2.get_mean(sacc.standard_types.galaxy_shear_cl_ee)
    for i in range(20):
        assert mu[i] == 0.1 * i


def test_io_maps_bpws():
    s = sacc.Sacc()

    n_ell = 10
    d_ell = 100
    n_ell_large = n_ell * d_ell
    ell = np.linspace(2, 1000, n_ell)
    c_ell = 1./(ell+1)**3
    beam = np.exp(-0.1 * ell * (ell+1))
    nu = np.linspace(30., 60., 100)
    bandpass = np.ones(100)
    z = np.arange(0., 1.0, 0.01)
    nz = (z-0.5)**2/0.1**2

    # Tracer
    s.add_tracer('NZ', 'gc', z, nz)
    s.add_tracer('NuMap', 'cmbp', 2, nu, bandpass, ell, beam)
    s.add_tracer('maP', 'sz', 0, ell, beam)

    # Window
    ells_large = np.arange(n_ell_large)
    window_single = np.zeros([n_ell, n_ell_large])
    for i in range(n_ell):
        window_single[i, i * d_ell: (i + 1) * d_ell] = 1.
    wins = sacc.BandpowerWindow(ells_large, window_single.T)

    s.add_ell_cl('cl_00', 'gc', 'gc', ell, c_ell, window=wins)
    s.add_ell_cl('cl_0e', 'gc', 'cmbp', ell, c_ell, window=wins)
    s.add_ell_cl('cl_00', 'gc', 'sz', ell, c_ell, window=wins)

    with tempfile.TemporaryDirectory() as tmpdir:
        filename = os.path.join(tmpdir, 'test.sacc')
        s.save_fits(filename)
        s2 = sacc.Sacc.load_fits(filename)

    assert len(s2) == 30
    l, cl, ind = s2.get_ell_cl('cl_00', 'gc', 'sz',
                               return_ind=True)
    w = s2.get_bandpower_windows(ind)
    assert np.all(cl == c_ell)
    assert w.weight.shape == (n_ell_large, n_ell)


def test_rename_tracer():
    s = get_filled_sacc()

    tracer_comb = s.get_tracer_combinations()

    s.rename_tracer('source_0', 'src_0')

    # Check that the name attribute of the tracer has changed
    tr = s.get_tracer('src_0')
    assert tr.name == 'src_0'

    # Check that the data tracer combinations have been updated
    tracer_comb_new = s.get_tracer_combinations()
    for trs, trs_new in zip(tracer_comb, tracer_comb_new):
        trs_n = []
        for tri in trs:
            if 'source_0' == tri:
                tri = 'src_0'
            trs_n.append(tri)

        assert tuple(trs_n) == trs_new

    # Check it is permanent
    with tempfile.TemporaryDirectory() as tmpdir:
        filename = os.path.join(tmpdir, 'test.sacc')
        s.save_fits(filename)
        s2 = sacc.Sacc.load_fits(filename)

    assert ('source_0' not in s2.tracers) and ('src_0' in s2.tracers)
    assert sorted(tracer_comb_new) == sorted(s2.get_tracer_combinations())
    # Make sure the tracers are ordered the same way when comparing the data
    # points
    s.to_canonical_order()
    s2.to_canonical_order()
    assert np.all(s.mean == s2.mean)
    assert np.all(s.indices(tracers=('src_0', 'source_1')) ==
                  s2.indices(tracers=('src_0', 'source_1')))
    assert np.all(s.indices(tracers=('source_1', 'source_1', 'src_0')) ==
                  s2.indices(tracers=('source_1', 'source_1', 'src_0')))


def test_qpnz_tracer():
    md1 = {'potato': 'if_necessary', 'answer': 42, 'height': 1.83}
    md2 = {'potato': 'never'}
    z = np.linspace(0., 1., 101)

    nz_qp_interp = qp.Ensemble(qp.interp, data=dict(xvals=z, yvals=np.ones(shape=(1, 101))))
    nz_qp_hist = qp.Ensemble(qp.hist, data=dict(bins=z, pdfs=np.ones(shape=(1, 100))))

    T1 = sacc.BaseTracer.make('QPNZ', 'tracer1', nz_qp_interp,
                              quantity='galaxy_density',
                              metadata=md1)
    T2 = sacc.BaseTracer.make('QPNZ', 'tracer2', nz_qp_hist,
                              quantity='galaxy_shear',
                              metadata=md2)
    assert T1.metadata == md1
    assert T2.metadata == md2

    tables = sacc.BaseTracer.to_tables([T1, T2])
    D = sacc.BaseTracer.from_tables(tables)

    T1a = D['tracer1']
    T2a = D['tracer2']
    assert T1a.metadata == md1
    assert T2a.metadata == md2


def test_io_qp():
    s = sacc.Sacc()

    # Tracer
    z = np.linspace(0., 1.0, 101)
    nz = np.expand_dims((z-0.5)**2/0.1**2, 0)
    ens = qp.Ensemble(qp.interp, data=dict(xvals=z, yvals=nz))
    ens.set_ancil(dict(modes = ens.mode(z)))
    s.add_tracer('QpnZ', 'source_0', ens)

    for i in range(20):
        ee = 0.1 * i
        tracers = ('source_0', 'source_0')
        s.add_data_point(sacc.standard_types.galaxy_shear_cl_ee,
                         tracers, ee, ell=10.0*i)

    with tempfile.TemporaryDirectory() as tmpdir:
        filename = os.path.join(tmpdir, 'test.sacc')
        s.save_fits(filename)
        s2 = sacc.Sacc.load_fits(filename)

    assert len(s2) == 20
    mu = s2.get_mean(sacc.standard_types.galaxy_shear_cl_ee)
    for i in range(20):
        assert mu[i] == 0.1 * i



"""
Tests for function supporting SACC.

"""

from sacc.tracers import (
    BinZTracer,
    BinLogMTracer,
    BinRichnessTracer,
    BinRadiusTracer,
    SurveyTracer,
)


def test_make_binztracer():
    tracer = BinZTracer.make("bin_z", name="fred", lower=0.5, upper=1.0)
    assert isinstance(tracer, BinZTracer)
    assert tracer.quantity == "generic"
    assert tracer.name == "fred"
    assert tracer.lower == 0.5
    assert tracer.upper == 1.0


def test_binztracer_equality():
    a = BinZTracer.make("bin_z", name="fred", lower=0.5, upper=1.0)
    b = BinZTracer.make("bin_z", name="fred", lower=0.5, upper=1.0)
    c = BinZTracer.make("bin_z", name="wilma", lower=0.5, upper=1.0)
    d = BinZTracer.make("bin_z", name="fred", lower=0.6, upper=1.0)
    e = BinZTracer.make("bin_z", name="fred", lower=0.5, upper=1.1)
    assert a == b
    assert a != "fred"
    assert a != c
    assert a != d
    assert a != e


def test_binztracer_tables():
    a = BinZTracer.make("bin_z", name="fred", lower=0.5, upper=1.0)
    b = BinZTracer.make("bin_z", name="wilma", lower=1.0, upper=1.5)
    tables = BinZTracer.to_tables([a, b])
    assert len(tables) == 1  # all BinZTracers are written to a single table

    d = BinZTracer.from_tables(tables)
    assert len(d) == 2  # this list of tables recovers both BinZTracers
    assert d["fred"] == a
    assert d["wilma"] == b


def test_make_binlogmtracer():
    tracer = BinLogMTracer.make("bin_logm", name="fred", lower=13.0, upper=15.0)
    assert isinstance(tracer, BinLogMTracer)
    assert tracer.quantity == "generic"
    assert tracer.name == "fred"
    assert tracer.lower == 13.0
    assert tracer.upper == 15.0


def test_binlogmtracer_equality():
    a = BinLogMTracer.make("bin_logm", name="fred", lower=13.0, upper=15.0)
    b = BinLogMTracer.make("bin_logm", name="fred", lower=13.0, upper=15.0)
    c = BinLogMTracer.make("bin_logm", name="wilma", lower=13.0, upper=15.0)
    d = BinLogMTracer.make("bin_logm", name="fred", lower=14.0, upper=15.0)
    e = BinLogMTracer.make("bin_logm", name="fred", lower=13.0, upper=15.1)
    assert a == b
    assert a != "fred"
    assert a != c
    assert a != d
    assert a != e


def test_binlogmtracer_tables():
    a = BinLogMTracer.make("bin_logm", name="fred", lower=13.0, upper=15.0)
    b = BinLogMTracer.make("bin_logm", name="wilma", lower=14.0, upper=15.5)
    tables = BinLogMTracer.to_tables([a, b])
    assert len(tables) == 1

    d = BinLogMTracer.from_tables(tables)
    assert len(d) == 2
    assert d["fred"] == a
    assert d["wilma"] == b


def test_make_binrichness_tracer():
    tracer = BinRichnessTracer.make(
        "bin_richness",
        name="barney",
        lower=0.25,
        upper=1.0,
    )
    assert isinstance(tracer, BinRichnessTracer)
    assert tracer.quantity == "generic"
    assert tracer.name == "barney"
    assert tracer.upper == 1.0
    assert tracer.lower == 0.25


def test_binrichnesstracer_equality():
    a = BinRichnessTracer.make("bin_richness", name="fred", lower=0.5, upper=1.0)
    b = BinRichnessTracer.make("bin_richness", name="fred", lower=0.5, upper=1.0)
    c = BinRichnessTracer.make("bin_richness", name="wilma", lower=0.5, upper=1.0)
    d = BinRichnessTracer.make("bin_richness", name="fred", lower=0.6, upper=1.0)
    e = BinRichnessTracer.make("bin_richness", name="fred", lower=0.5, upper=1.1)
    assert a == b
    assert a != "fred"
    assert a != c
    assert a != d
    assert a != e


def test_binrichnesstracer_tables():
    a = BinRichnessTracer.make("bin_richness", name="barney", lower=0.0, upper=0.5)
    b = BinRichnessTracer.make("bin_richness", name="betty", lower=1.25, upper=2.0)
    tables = BinRichnessTracer.to_tables([a, b])
    assert len(tables) == 1
    d = BinRichnessTracer.from_tables(tables)
    assert len(d) == 2  # this list of tables recovers both BinRichnessTracers
    assert d["barney"] == a
    assert d["betty"] == b


def test_make_binradiustracer():
    tracer = BinRadiusTracer.make(
        "bin_radius", name="pebbles", lower=1.0, center=2.0, upper=3.0
    )
    assert isinstance(tracer, BinRadiusTracer)
    assert tracer.quantity == "generic"
    assert tracer.name == "pebbles"
    assert tracer.lower == 1.0
    assert tracer.center == 2.0
    assert tracer.upper == 3.0

def test_binradiustracer_equality():
    a = BinRadiusTracer.make(
        "bin_radius", name="fred", lower=0.5, center=0.75, upper=1.0
    )
    b = BinRadiusTracer.make(
        "bin_radius", name="fred", lower=0.5, center=0.75, upper=1.0
    )
    c = BinRadiusTracer.make(
        "bin_radius", name="wilma", lower=0.5, center=0.75, upper=1.0
    )
    d = BinRadiusTracer.make(
        "bin_radius", name="fred", lower=0.6, center=0.75, upper=1.0
    )
    e = BinRadiusTracer.make(
        "bin_radius", name="fred", lower=0.5, center=0.8, upper=1.0
    )
    f = BinRadiusTracer.make(
        "bin_radius", name="fred", lower=0.5, center=0.75, upper=1.1
    )
    assert a == b
    assert a != "fred"
    assert a != c
    assert a != d
    assert a != e
    assert a != f


def test_binradiustracer_tables():
    a = BinRadiusTracer.make(
        "bin_radius", name="pebbles", lower=1.0, center=2.0, upper=3.0
    )
    b = BinRadiusTracer.make(
        "bin_radius", name="bambam", lower=3.0, center=4.0, upper=5.0
    )
    tables = BinRadiusTracer.to_tables([a, b])
    assert len(tables) == 1
    d = BinRadiusTracer.from_tables(tables)
    assert len(d) == 2
    assert d["pebbles"] == a
    assert d["bambam"] == b


def test_make_surveytracer():
    tracer = SurveyTracer.make("survey", name="bullwinkle", sky_area=1.0)
    assert isinstance(tracer, SurveyTracer)
    assert tracer.quantity == "generic"
    assert tracer.name == "bullwinkle"
    assert tracer.sky_area == 1.0


def test_surveytracer_equality():
    a = SurveyTracer.make("survey", name="bullwinkle", sky_area=1.0)
    b = SurveyTracer.make("survey", name="bullwinkle", sky_area=1.0)
    c = SurveyTracer.make("survey", name="rocky", sky_area=1.0)
    d = SurveyTracer.make("survey", name="boris", sky_area=2.0)

    assert a == b
    assert a != "bullwinkle"
    assert a != c
    assert a != d


def test_surveytracer_tables():
    a = SurveyTracer.make("survey", name="bullwinkle", sky_area=1.0)
    b = SurveyTracer.make("survey", name="rocky", sky_area=2.0)
    tables = SurveyTracer.to_tables([a, b])
    assert len(tables) == 1
    d = SurveyTracer.from_tables(tables)
    assert len(d) == 2
    assert d["bullwinkle"] == a
    assert d["rocky"] == b



import numpy as np
from astropy.table import Table
from .utils import (Namespace, hide_null_values,
                    remove_dict_null_values, unique_list)
import warnings

standard_quantities = Namespace('galaxy_shear',
                                'galaxy_density',
                                'galaxy_convergence',
                                'cluster_density',
                                'cmb_temperature',
                                'cmb_polarization',
                                'cmb_convergence',
                                'cmb_tSZ',
                                'cmb_kSZ',
                                'cluster_mass_count_wl',
                                'cluster_mass_count_xray',
                                'cluster_mass_count_tSZ',
                                'generic')


class BaseTracer:
    """
    A class representing some kind of tracer of astronomical objects.

    Generically, SACC data points correspond to some combination of tracers
    for example, tomographic two-point data has two tracers for each data
    point, indicating the n(z) for the corresponding tomographic bin.

    All Tracer objects have at least a name attribute.  Different
    subclassses have other requirements.  For example, n(z) tracers
    require z and n(z) arrays.

    In general you don't need to create tracer objects yourself -
    the Sacc.add_tracer method will construct them for you.
    """
    _tracer_classes = {}

    def __init__(self, name, **kwargs):
        # We encourage people to use existing quantity names, and issue a
        # warning if they do not to prod them in the right direction.
        quantity = kwargs.pop('quantity', 'generic')
        if quantity not in standard_quantities:
            warnings.warn(f"Unknown quantity {quantity}. "
                          "If possible use a pre-defined quantity, or "
                          "add to the list.")
        self.name = name
        self.quantity = quantity
        self.metadata = kwargs.pop('metadata', {})

    def __init_subclass__(cls, tracer_type):
        cls._tracer_classes[tracer_type.lower()] = cls
        cls.tracer_type = tracer_type

    @classmethod
    def make(cls, tracer_type, name, *args, **kwargs):
        """
        Select a Tracer subclass based on tracer_type
        and instantiate in instance of it with the remaining
        arguments.

        Parameters
        ----------
        tracer_type: str
            Must correspond to the tracer_type of a subclass

        name: str
            The name for this specific tracer.

        Returns
        -------
        instance: Tracer object
            An instance of a Tracer subclass
        """
        subclass = cls._tracer_classes[tracer_type.lower()]
        obj = subclass(name, *args, **kwargs)
        return obj

    @classmethod
    def to_tables(cls, instance_list):
        """Convert a list of tracers to a list of astropy tables

        This is used when saving data to a file.

        This class method converts a list of tracers, each of which
        can instances of any subclass of BaseTracer, and turns them
        into a list of astropy tables, ready to be saved to disk.

        Some tracers generate a single table for all of the
        different instances, and others generate one table per
        instance.

        Parameters
        ----------
        instance_list: list
            List of tracer instances

        Returns
        -------
        tables: list
            List of astropy tables
        """
        tables = []
        for name, subcls in cls._tracer_classes.items():
            tracers = [t for t in instance_list if type(t) == subcls]
            # If the list is empty, we don't want to append any tables.
            if tracers:
                tables += subcls.to_tables(tracers)
        return tables

    @classmethod
    def from_tables(cls, table_list):
        """Convert a list of astropy tables into a dictionary of tracers

        This is used when loading data from a file.

        This class method takes a list of tracers, such as those
        read from a file, and converts them into a list of instances.

        It is not quite the inverse of the to_tables method, since it
        returns a dict instead of a list.

        Subclasses overrides of this method do the actual work, but
        should *NOT* call this parent base method.

        Parameters
        ----------
        table_list: list
            List of astropy tables

        Returns
        -------
        tracers: dict
            Dict mapping string names to tracer objects.
        """
        tracers = {}
        # Figure out the different subclasses that are present
        subclass_names = unique_list(table.meta['SACCCLSS'].lower()
                                     for table in table_list)
        subclasses = [cls._tracer_classes[name]
                      for name in subclass_names]

        # For each subclass find the tables representing that subclass.
        # We do it like this because we might want to represent one tracer with
        # multiple tables, or one table can have multiple tracers -
        # it depends on the tracers class and how complicated it is.
        for name, subcls in zip(subclass_names, subclasses):
            subcls_table_list = [table for table in table_list
                                 if table.meta['SACCCLSS'].lower() == name]
            # and ask the subclass to read from those tables.
            tracers.update(subcls.from_tables(subcls_table_list))
        return tracers


class MiscTracer(BaseTracer, tracer_type='Misc'):
    """A Tracer type for miscellaneous other data points.

    MiscTracers do not have any attributes except for their
    name, so can be used for tagging external data, for example.

    Parameters
    ----------
    name: str
        The name of the tracer
    """

    def __init__(self, name, **kwargs):
        super().__init__(name, **kwargs)

    @classmethod
    def to_tables(cls, instance_list):
        """Convert a list of MiscTracer instances to a astropy tables.

        This is used when saving data to file.

        All the instances are converted to a single table, which is
        returned in a list with one element so that it can be used
        in combination with the parent.

        You can use the parent class to_tables class method to convert
        a mixed list of different tracer types.

        You shouldn't generally need to call this method directly.

        Parameters
        ----------
        instance_list: list
            list of MiscTracer objects

        Returns
        -------
        tables: list
            List containing one astropy table
        """
        metadata_cols = set()
        for obj in instance_list:
            metadata_cols.update(obj.metadata.keys())
        metadata_cols = list(metadata_cols)

        cols = [[obj.name for obj in instance_list],
                [obj.quantity for obj in instance_list]]
        for name in metadata_cols:
            cols.append([obj.metadata.get(name) for obj in instance_list])

        table = Table(data=cols,
                      names=['name', 'quantity'] + metadata_cols)
        table.meta['SACCTYPE'] = 'tracer'
        table.meta['SACCCLSS'] = cls.tracer_type
        table.meta['EXTNAME'] = f'tracer:{cls.tracer_type}'
        hide_null_values(table)
        return [table]

    @classmethod
    def from_tables(cls, table_list):
        """Convert a list of astropy table into a dictionary of MiscTracer instances.

        In general table_list should have a single element in, since all the
        MiscTracers are stored in a single table during to_tables

        Parameters
        ----------
        table_list: List[astropy.table.Table]

        Returns
        -------
        tracers: Dict[str: MiscTracer]
        """
        tracers = {}

        for table in table_list:
            metadata_cols = [col for col in table.colnames
                             if col not in ['name', 'quantity']]

            for row in table:
                name = row['name']
                quantity = row['quantity']
                metadata = {key: row[key] for key in metadata_cols}
                remove_dict_null_values(metadata)
                tracers[name] = cls(name, quantity=quantity, metadata=metadata)
        return tracers


class MapTracer(BaseTracer, tracer_type='Map'):
    """
    A Tracer type for a sky map.

    Takes at least two arguments, defining the map beam.

    Parameters
    ----------
    name: str
        The name for this specific tracer object.
    ell: array
         Array of multipole values at which the beam is defined.
    beam: array
         Beam multipoles at each value of ell.
    beam_extra: array
         Other beam-related arrays
         (e.g. uncertainties, principal components,
         alternative measurements, whatever).
    map_unit: str
         Map units (e.g. 'uK_CMB'). 'none' by default.
    """

    def __init__(self, name, spin, ell, beam,
                 beam_extra=None, map_unit='none', **kwargs):
        super().__init__(name, **kwargs)
        self.spin = spin
        self.map_unit = map_unit
        self.ell = np.array(ell)
        self.beam = np.array(beam)
        self.beam_extra = {} if beam_extra is None else beam_extra

    @classmethod
    def to_tables(cls, instance_list):
        tables = []
        for tracer in instance_list:
            # Beams
            names = ['ell', 'beam']
            cols = [tracer.ell, tracer.beam]
            for beam_id, col in tracer.beam_extra.items():
                names.append(str(beam_id))
                cols.append(col)
            table = Table(data=cols, names=names)
            table.meta['SACCTYPE'] = 'tracer'
            table.meta['SACCCLSS'] = cls.tracer_type
            table.meta['SACCNAME'] = tracer.name
            table.meta['SACCQTTY'] = tracer.quantity
            extname = f'tracer:{cls.tracer_type}:{tracer.name}:beam'
            table.meta['EXTNAME'] = extname
            table.meta['MAP_UNIT'] = tracer.map_unit
            table.meta['SPIN'] = tracer.spin
            for key, value in tracer.metadata.items():
                table.meta['META_'+key] = value
            remove_dict_null_values(table.meta)
            tables.append(table)
        return tables

    @classmethod
    def from_tables(cls, table_list):
        tracers = {}

        # Collect beam and bandpass tables describing the same tracer
        tr_tables = {}
        for table in table_list:
            # Read name and table type
            name = table.meta['SACCNAME']
            quantity = table.meta.get('SACCQTTY', 'generic')
            tabtyp = table.meta['EXTNAME'].split(':')[-1]
            if tabtyp not in ['beam']:
                raise KeyError("Unknown table type " + table.meta['EXTNAME'])

            # If not present yet, create new tracer entry
            if name not in tr_tables:
                tr_tables[name] = {}
            # Add table
            tr_tables[name][tabtyp] = table

        # Now loop through different tracers and build them from their tables
        for n, dt in tr_tables.items():
            quantity = []
            metadata = {}
            map_unit = 'none'
            ell = []
            beam = []
            beam_extra = {}
            spin = 0

            if 'beam' in dt:
                table = dt['beam']
                name = table.meta['SACCNAME']
                quantity = table.meta.get('SACCQTTY', 'generic')
                ell = table['ell']
                beam = table['beam']
                for col in table.columns.values():
                    if col.name not in ['ell', 'beam']:
                        beam_extra[col.name] = col.data
                map_unit = table.meta['MAP_UNIT']
                spin = table.meta['SPIN']
                for key, value in table.meta.items():
                    if key.startswith("META_"):
                        metadata[key[5:]] = value

            tracers[name] = cls(name, spin, ell, beam,
                                quantity=quantity, beam_extra=beam_extra,
                                map_unit=map_unit, metadata=metadata)
        return tracers


class NuMapTracer(BaseTracer, tracer_type='NuMap'):
    """
    A Tracer type for a sky map at a given frequency.

    Takes at least four arguments, defining the bandpass and beam.

    Parameters
    ----------
    name: str
        The name for this specific tracer, e.g. a frequency band
        identifier.
    spin: int
        Spin for this observable. Either 0 (e.g. intensity)
        or 2 (e.g. polarization).
    nu: array
         Array of frequencies.
    bandpass: array
         Bandpass transmission.
    bandpass_extra: array
         Other bandpass-related arrays
         (e.g. uncertainties, principal components,
         alternative measurements, whatever).
    ell: array
         Array of multipole values at which the beam is defined.
    beam: array
         Beam.
    beam_extra: array
         Other beam-related arrays
         (e.g. uncertainties, principal components,
         alternative measurements, whatever).
    nu_unit: str
         Frequency units ('GHz' by default).
    map_unit: str
         Map units (e.g. 'uK_CMB'). 'none' by default.
    """

    def __init__(self, name, spin, nu, bandpass,
                 ell, beam, bandpass_extra=None,
                 beam_extra=None, nu_unit='GHz',
                 map_unit='none', **kwargs):
        super().__init__(name, **kwargs)
        self.spin = spin
        self.nu = np.array(nu)
        self.nu_unit = nu_unit
        self.map_unit = map_unit
        self.bandpass = np.array(bandpass)
        self.bandpass_extra = {} if bandpass_extra is None else bandpass_extra
        self.ell = np.array(ell)
        self.beam = np.array(beam)
        self.beam_extra = {} if beam_extra is None else beam_extra

    @classmethod
    def to_tables(cls, instance_list):
        tables = []
        for tracer in instance_list:
            # Bandpasses
            names = ['nu', 'bandpass']
            cols = [tracer.nu, tracer.bandpass]
            for bandpass_id, col in tracer.bandpass_extra.items():
                names.append(str(bandpass_id))
                cols.append(col)
            table = Table(data=cols, names=names)
            table.meta['SACCTYPE'] = 'tracer'
            table.meta['SACCCLSS'] = cls.tracer_type
            table.meta['SACCNAME'] = tracer.name
            table.meta['SACCQTTY'] = tracer.quantity
            extname = f'tracer:{cls.tracer_type}:{tracer.name}:bandpass'
            table.meta['EXTNAME'] = extname
            table.meta['NU_UNIT'] = tracer.nu_unit
            table.meta['SPIN'] = tracer.spin
            for key, value in tracer.metadata.items():
                table.meta['META_'+key] = value
            remove_dict_null_values(table.meta)
            tables.append(table)

            # Beams
            names = ['ell', 'beam']
            cols = [tracer.ell, tracer.beam]
            for beam_id, col in tracer.beam_extra.items():
                names.append(str(beam_id))
                cols.append(col)
            table = Table(data=cols, names=names)
            table.meta['SACCTYPE'] = 'tracer'
            table.meta['SACCCLSS'] = cls.tracer_type
            table.meta['SACCNAME'] = tracer.name
            table.meta['SACCQTTY'] = tracer.quantity
            extname = f'tracer:{cls.tracer_type}:{tracer.name}:beam'
            table.meta['EXTNAME'] = extname
            table.meta['MAP_UNIT'] = tracer.map_unit
            table.meta['SPIN'] = tracer.spin
            for key, value in tracer.metadata.items():
                table.meta['META_'+key] = value
            remove_dict_null_values(table.meta)
            tables.append(table)
        return tables

    @classmethod
    def from_tables(cls, table_list):
        tracers = {}

        # Collect beam and bandpass tables describing the same tracer
        tr_tables = {}
        for table in table_list:
            # Read name and table type
            name = table.meta['SACCNAME']
            quantity = table.meta.get('SACCQTTY', 'generic')
            tabtyp = table.meta['EXTNAME'].split(':')[-1]
            if tabtyp not in ['bandpass', 'beam']:
                raise KeyError("Unknown table type " + table.meta['EXTNAME'])

            # If not present yet, create new tracer entry
            if name not in tr_tables:
                tr_tables[name] = {}
            # Add table
            tr_tables[name][tabtyp] = table

        # Now loop through different tracers and build them from their tables
        for n, dt in tr_tables.items():
            quantity = []
            metadata = {}
            nu = []
            bandpass = []
            bandpass_extra = {}
            nu_unit = 'GHz'
            map_unit = 'none'
            ell = []
            beam = []
            beam_extra = {}
            spin = 0

            if 'bandpass' in dt:
                table = dt['bandpass']
                name = table.meta['SACCNAME']
                quantity = table.meta.get('SACCQTTY', 'generic')
                nu = table['nu']
                bandpass = table['bandpass']
                for col in table.columns.values():
                    if col.name not in ['nu', 'bandpass']:
                        bandpass_extra[col.name] = col.data
                nu_unit = table.meta['NU_UNIT']
                spin = table.meta['SPIN']
                for key, value in table.meta.items():
                    if key.startswith("META_"):
                        metadata[key[5:]] = value

            if 'beam' in dt:
                table = dt['beam']
                name = table.meta['SACCNAME']
                quantity = table.meta.get('SACCQTTY', 'generic')
                ell = table['ell']
                beam = table['beam']
                for col in table.columns.values():
                    if col.name not in ['ell', 'beam']:
                        beam_extra[col.name] = col.data
                map_unit = table.meta['MAP_UNIT']
                spin = table.meta['SPIN']
                for key, value in table.meta.items():
                    if key.startswith("META_"):
                        metadata[key[5:]] = value

            tracers[name] = cls(name, spin,
                                nu, bandpass,
                                ell, beam,
                                quantity=quantity,
                                bandpass_extra=bandpass_extra,
                                beam_extra=beam_extra,
                                map_unit=map_unit,
                                nu_unit=nu_unit,
                                metadata=metadata)
        return tracers


class NZTracer(BaseTracer, tracer_type='NZ'):
    """
    A Tracer type for tomographic n(z) data.

    Takes two arguments arrays of z and n(z)

    Parameters
    ----------
    name: str
        The name for this specific tracer, e.g. a
        tomographic bin identifier.

    z: array
        Redshift sample values

    nz: array
        Number density n(z) at redshift sample points.

    extra_columns: dict[str: array] or dict[int: array]
        Additional estimates of the same n(z), by name
    """

    def __init__(self, name, z, nz,
                 extra_columns=None, **kwargs):
        """
        Create a tracer corresponding to a distribution in redshift n(z),
        for example of galaxies.

        Parameters
        ----------
        name: str
            The name for this specific tracer, e.g. a
            tomographic bin identifier.

        z: array
            Redshift sample values

        nz: array
            Number density n(z) at redshift sample points.

        extra_columns: dict[str:array]
            Optional, default=None.  Additional realizations or
            estimates of the same n(z), by name.

        Returns
        -------
        instance: NZTracer object
            An instance of this class
        """
        super().__init__(name, **kwargs)
        self.z = np.array(z)
        self.nz = np.array(nz)
        self.extra_columns = {} if extra_columns is None else extra_columns

    @classmethod
    def to_tables(cls, instance_list):
        """Convert a list of NZTracers to a list of astropy tables

        This is used when saving data to a file.
        One table is generated per tracer.

        Parameters
        ----------
        instance_list: list
            List of tracer instances

        Returns
        -------
        tables: list
            List of astropy tables
        """
        tables = []
        for tracer in instance_list:
            names = ['z', 'nz']
            cols = [tracer.z, tracer.nz]
            for nz_id, col in tracer.extra_columns.items():
                names.append(str(nz_id))
                cols.append(col)
            table = Table(data=cols, names=names)
            table.meta['SACCTYPE'] = 'tracer'
            table.meta['SACCCLSS'] = cls.tracer_type
            table.meta['SACCNAME'] = tracer.name
            table.meta['SACCQTTY'] = tracer.quantity
            table.meta['EXTNAME'] = f'tracer:{cls.tracer_type}:{tracer.name}'
            for key, value in tracer.metadata.items():
                table.meta['META_'+key] = value
            remove_dict_null_values(table.meta)
            tables.append(table)
        return tables

    @classmethod
    def from_tables(cls, table_list):
        """Convert an astropy table into a dictionary of tracers

        This is used when loading data from a file.
        A single tracer object is read from the table.

        Parameters
        ----------
        table_list: list[astropy.table.Table]
            Must contain the appropriate data, for example as saved
            by to_table.

        Returns
        -------
        tracers: dict
            Dict mapping string names to tracer objects.
            Only contains one key/value pair for the one tracer.
        """
        tracers = {}
        for table in table_list:
            name = table.meta['SACCNAME']
            quantity = table.meta.get('SACCQTTY', 'generic')
            z = table['z']
            nz = table['nz']
            extra_columns = {}
            for col in table.columns.values():
                if col.name not in ['z', 'nz']:
                    extra_columns[col.name] = col.data

            metadata = {}
            for key, value in table.meta.items():
                if key.startswith("META_"):
                    metadata[key[5:]] = value
            tracers[name] = cls(name, z, nz,
                                quantity=quantity,
                                extra_columns=extra_columns,
                                metadata=metadata)
        return tracers


class QPNZTracer(BaseTracer, tracer_type='QPNZ'):
    """
    A Tracer type for tomographic n(z) data preresented as a `qp.Ensemble`

    Takes a `qp.Ensemble`

    Parameters
    ----------
    name: str
        The name for this specific tracer, e.g. a
        tomographic bin identifier.

    ensemble: qp.Ensemble
        The qp.ensemble in questions
    """

    def __init__(self, name, ens, **kwargs):
        """
        Create a tracer corresponding to a distribution in redshift n(z),
        for example of galaxies.

        Parameters
        ----------
        name: str
            The name for this specific tracer, e.g. a
            tomographic bin identifier.

        ensemble: qp.Ensemble
            The qp.ensemble in questions

        Returns
        -------
        instance: NZTracer object
            An instance of this class
        """
        super().__init__(name, **kwargs)
        self.ensemble = ens

    @classmethod
    def to_tables(cls, instance_list):
        """Convert a list of NZTracers to a list of astropy tables

        This is used when saving data to a file.
        Two or three tables are generated per tracer.

        Parameters
        ----------
        instance_list: list
            List of tracer instances

        Returns
        -------
        tables: list
            List of astropy tables
        """
        from tables_io.convUtils import convertToApTables

        tables = []

        for tracer in instance_list:
            table_dict = tracer.ensemble.build_tables()
            ap_tables = convertToApTables(table_dict)
            data_table = ap_tables['data']
            meta_table = ap_tables['meta']
            ancil_table = ap_tables.get('ancil', None)
            meta_table.meta['SACCTYPE'] = 'tracer'
            meta_table.meta['SACCCLSS'] = cls.tracer_type
            meta_table.meta['SACCNAME'] = tracer.name
            meta_table.meta['SACCQTTY'] = tracer.quantity
            meta_table.meta['EXTNAME'] = f'tracer:{cls.tracer_type}:{tracer.name}:meta'

            data_table.meta['SACCTYPE'] = 'tracer'
            data_table.meta['SACCCLSS'] = cls.tracer_type
            data_table.meta['SACCNAME'] = tracer.name
            data_table.meta['SACCQTTY'] = tracer.quantity
            data_table.meta['EXTNAME'] = f'tracer:{cls.tracer_type}:{tracer.name}:data'

            for kk, vv in tracer.metadata.items():
                meta_table.meta['META_'+kk] = vv
            tables.append(data_table)
            tables.append(meta_table)
            if ancil_table:
                ancil_table.meta['SACCTYPE'] = 'tracer'
                ancil_table.meta['SACCCLSS'] = cls.tracer_type
                ancil_table.meta['SACCNAME'] = tracer.name
                ancil_table.meta['SACCQTTY'] = tracer.quantity
                ancil_table.meta['EXTNAME'] = f'tracer:{cls.tracer_type}:{tracer.name}:ancil'

                tables.append(ancil_table)
        return tables

    @classmethod
    def from_tables(cls, table_list):
        """Convert an astropy table into a dictionary of tracers

        This is used when loading data from a file.
        A single tracer object is read from the table.

        Parameters
        ----------
        table_list: list[astropy.table.Table]
            Must contain the appropriate data, for example as saved
            by to_table.

        Returns
        -------
        tracers: dict
            Dict mapping string names to tracer objects.
            Only contains one key/value pair for the one tracer.
        """
        import qp
        tracers = {}
        sorted_dict = {}
        for table_ in table_list:
            tokens = table_.meta['EXTNAME'].split(':')
            table_key = f'{tokens[0]}:{tokens[1]}:{tokens[2]}'
            table_type = f'{tokens[3]}'
            if table_key not in sorted_dict:
                sorted_dict[table_key] = {table_type: table_}
            else:
                sorted_dict[table_key][table_type] = table_

        for key, val in sorted_dict.items():
            meta_table = val['meta']
            ensemble = qp.from_tables(val)
            name = meta_table.meta['SACCNAME']
            quantity = meta_table.meta.get('SACCQTTY', 'generic')
            ensemble = qp.from_tables(val)
            metadata = {}
            for key, value in meta_table.meta.items():
                if key.startswith("META_"):
                    metadata[key[5:]] = value
            tracers[name] = cls(name, ensemble,
                                quantity=quantity,
                                metadata=metadata)
        return tracers


class BinZTracer(BaseTracer, tracer_type="bin_z"):  # type: ignore
    """A tracer for a single redshift bin. The tracer shall
    be used for binned data where we want a desired quantity
    per interval of redshift, such that we only need the data
    for a given interval instead of at individual redshifts."""

    def __init__(self, name: str, lower: float, upper: float, **kwargs):
        """
        Create a tracer corresponding to a single redshift bin.

        :param name: The name of the tracer
        :param lower: The lower bound of the redshift bin
        :param upper: The upper bound of the redshift bin
        """
        super().__init__(name, **kwargs)
        self.lower = lower
        self.upper = upper

    def __eq__(self, other) -> bool:
        """Test for equality.  If :python:`other` is not a
        :python:`BinZTracer`, then it is not equal to :python:`self`.
        Otherwise, they are equal if names, and the z-range of the bins,
        are equal."""
        if not isinstance(other, BinZTracer):
            return False
        return (
            self.name == other.name
            and self.lower == other.lower
            and self.upper == other.upper
        )

    @classmethod
    def to_tables(cls, instance_list):
        """Convert a list of BinZTracers to a single astropy table

        This is used when saving data to a file.
        One table is generated with the information for all the tracers.

        :param instance_list: List of tracer instances
        :return: List with a single astropy table
        """

        names = ["name", "quantity", "lower", "upper"]

        cols = [
            [obj.name for obj in instance_list],
            [obj.quantity for obj in instance_list],
            [obj.lower for obj in instance_list],
            [obj.upper for obj in instance_list],
        ]

        table = Table(data=cols, names=names)
        table.meta["SACCTYPE"] = "tracer"
        table.meta["SACCCLSS"] = cls.tracer_type
        table.meta["EXTNAME"] = f"tracer:{cls.tracer_type}"
        return [table]

    @classmethod
    def from_tables(cls, table_list):
        """Convert an astropy table into a dictionary of tracers

        This is used when loading data from a file.
        One tracer object is created for each "row" in each table.

        :param table_list: List of astropy tables
        :return: Dictionary of tracers
        """
        tracers = {}

        for table in table_list:
            for row in table:
                name = row["name"]
                quantity = row["quantity"]
                lower = row["lower"]
                upper = row["upper"]
                tracers[name] = cls(name, quantity=quantity, lower=lower, upper=upper)
        return tracers

class BinLogMTracer(BaseTracer, tracer_type="bin_logM"):  # type: ignore
    """A tracer for a single log-mass bin. The tracer shall
    be used for binned data where we want a desired quantity
    per interval of log(mass), such that we only need the data
    for a given interval instead of at individual masses."""

    def __init__(self, name: str, lower: float, upper: float, **kwargs):
        """
        Create a tracer corresponding to a single log-mass bin.

        :param name: The name of the tracer
        :param lower: The lower bound of the log-mass bin
        :param upper: The upper bound of the log-mass bin
        """
        super().__init__(name, **kwargs)
        self.lower = lower
        self.upper = upper

    def __eq__(self, other) -> bool:
        """Test for equality.  If :python:`other` is not a
        :python:`BinLogMTracer`, then it is not equal to :python:`self`.
        Otherwise, they are equal if names, and the z-range of the bins,
        are equal."""
        if not isinstance(other, BinLogMTracer):
            return False
        return (
            self.name == other.name
            and self.lower == other.lower
            and self.upper == other.upper
        )

    @classmethod
    def to_tables(cls, instance_list):
        """Convert a list of BinLogMTracers to a single astropy table

        This is used when saving data to a file.
        One table is generated with the information for all the tracers.

        :param instance_list: List of tracer instances
        :return: List with a single astropy table
        """

        names = ["name", "quantity", "lower", "upper"]

        cols = [
            [obj.name for obj in instance_list],
            [obj.quantity for obj in instance_list],
            [obj.lower for obj in instance_list],
            [obj.upper for obj in instance_list],
        ]
        table = Table(data=cols, names=names)
        table.meta["SACCTYPE"] = "tracer"
        table.meta["SACCCLSS"] = cls.tracer_type
        table.meta["EXTNAME"] = f"tracer:{cls.tracer_type}"
        return [table]

    @classmethod
    def from_tables(cls, table_list):
        """Convert an astropy table into a dictionary of tracers

        This is used when loading data from a file.
        One tracer object is created for each "row" in each table.

        :param table_list: List of astropy tables
        :return: Dictionary of tracers
        """
        tracers = {}

        for table in table_list:
            for row in table:
                name = row["name"]
                quantity = row["quantity"]
                lower = row["lower"]
                upper = row["upper"]
                tracers[name] = cls(name, quantity=quantity, lower=lower, upper=upper)
        return tracers

         
class BinRichnessTracer(BaseTracer, tracer_type="bin_richness"):  # type: ignore
    """A tracer for a single richness bin. The tracer shall
    be used for binned data where we want a desired quantity
    per interval of log(richness), such that we only need the data
    for a given interval instead of at individual richness."""

    def __eq__(self, other) -> bool:
        """Test for equality. If :python:`other` is not a
        :python:`BinRichnessTracer`, then it is not equal to :python:`self`.
        Otherwise, they are equal if names and the richness-range of the
        bins, are equal."""
        if not isinstance(other, BinRichnessTracer):
            return False
        return (
            self.name == other.name
            and self.lower == other.lower
            and self.upper == other.upper
        )

    def __init__(self, name: str, lower: float, upper: float, **kwargs):
        """
        Create a tracer corresponding to a single richness bin.

        :param name: The name of the tracer
        :param lower: The lower bound of the richness bin in log10.
        :param upper: The upper bound of the richness bin in log10.
        """
        super().__init__(name, **kwargs)
        self.lower = lower
        self.upper = upper

    @classmethod
    def to_tables(cls, instance_list):
        """Convert a list of BinZTracers to a list of astropy tables

        This is used when saving data to a file.
        One table is generated with the information for all the tracers.

        :param instance_list: List of tracer instances
        :return: List with a single astropy table
        """
        names = ["name", "quantity", "lower", "upper"]

        cols = [
            [obj.name for obj in instance_list],
            [obj.quantity for obj in instance_list],
            [obj.lower for obj in instance_list],
            [obj.upper for obj in instance_list],
        ]

        table = Table(data=cols, names=names)
        table.meta["SACCTYPE"] = "tracer"
        table.meta["SACCCLSS"] = cls.tracer_type
        table.meta["EXTNAME"] = f"tracer:{cls.tracer_type}"
        return [table]

    @classmethod
    def from_tables(cls, table_list):
        """Convert an astropy table into a dictionary of tracers

        This is used when loading data from a file.
        One tracer object is created for each "row" in each table.

        :param table_list: List of astropy tables
        :return: Dictionary of tracers
        """
        tracers = {}

        for table in table_list:
            for row in table:
                name = row["name"]
                quantity = row["quantity"]
                lower = row["lower"]
                upper = row["upper"]
                tracers[name] = cls(
                    name,
                    quantity=quantity,
                    lower=lower,
                    upper=upper,
                )
        return tracers


class BinRadiusTracer(BaseTracer, tracer_type="bin_radius"):  # type: ignore
    """A tracer for a single radial bin, e.g. when dealing with cluster shear profiles.
    It gives the bin edges and the value of the bin "center". The latter would typically
    be returned by CLMM and correspond to the average radius of the galaxies in that
    radial bin. """

    def __eq__(self, other) -> bool:
        """Test for equality. If :python:`other` is not a
        :python:`BinRadiusTracer`, then it is not equal to :python:`self`.
        Otherwise, they are equal if names and the r-range and centers of the
        bins, are equal."""
        if not isinstance(other, BinRadiusTracer):
            return False
        return (
            self.name == other.name
            and self.lower == other.lower
            and self.center == other.center
            and self.upper == other.upper
        )

    def __init__(self, name: str, lower: float, upper: float, center: float, **kwargs):
        """
        Create a tracer corresponding to a single radial bin.

        :param name: The name of the tracer
        :param lower: The lower bound of the radius bin
        :param upper: The upper bound of the radius bin
        :param center: The value to use if a single point-estimate is needed.

        Note that :python:`center` need not be the midpoint between
        :python:`lower` and :python:`upper`'.
        """
        super().__init__(name, **kwargs)
        self.lower = lower
        self.upper = upper
        self.center = center

    @classmethod
    def to_tables(cls, instance_list):
        """Convert a list of BinRadiusTracers to a single astropy table

        This is used when saving data to a file.
        One table is generated with the information for all the tracers.

        :param instance_list: List of tracer instances
        :return: List with a single astropy table
        """

        names = ["name", "quantity", "lower", "upper", "center"]

        cols = [
            [obj.name for obj in instance_list],
            [obj.quantity for obj in instance_list],
            [obj.lower for obj in instance_list],
            [obj.upper for obj in instance_list],
            [obj.center for obj in instance_list],
        ]

        table = Table(data=cols, names=names)
        table.meta["SACCTYPE"] = "tracer"
        table.meta["SACCCLSS"] = cls.tracer_type
        table.meta["EXTNAME"] = f"tracer:{cls.tracer_type}"
        return [table]

    @classmethod
    def from_tables(cls, table_list):
        """Convert an astropy table into a dictionary of tracers

        This is used when loading data from a file.
        One tracer object is created for each "row" in each table.

        :param table_list: List of astropy tables
        :return: Dictionary of tracers
        """
        tracers = {}

        for table in table_list:
            for row in table:
                name = row["name"]
                quantity = row["quantity"]
                lower = row["lower"]
                upper = row["upper"]
                center = row["center"]
                tracers[name] = cls(
                    name,
                    quantity=quantity,
                    lower=lower,
                    upper=upper,
                    center=center,
                )
        return tracers

class SurveyTracer(BaseTracer, tracer_type="survey"):  # type: ignore
    """A tracer for the survey definition. It shall 
    be used to filter data related to a given survey
    and to provide the survey sky-area of analysis."""

    def __eq__(self, other) -> bool:
        """Test for equality. If :python:`other` is not a
        :python:`SurveyTracer`, then it is not equal to :python:`self`.
        Otherwise, they are equal if names and the sky-areas are equal."""
        if not isinstance(other, SurveyTracer):
            return False
        return self.name == other.name and self.sky_area == other.sky_area

    def __init__(self, name: str, sky_area: float, **kwargs):
        """
        Create a tracer corresponding to the survey definition.

        :param name: The name of the tracer
        :param sky_area: The survey's sky area in square degrees
        """
        super().__init__(name, **kwargs)
        self.sky_area = sky_area

    @classmethod
    def to_tables(cls, instance_list):
        """Convert a list of SurveyTracer to a list of astropy tables

        This is used when saving data to a file.
        One table is generated with the information for all the tracers.

        :param instance_list: List of tracer instances
        :return: List of astropy tables with one table
        """
        names = ["name", "quantity", "sky_area"]

        cols = [
            [obj.name for obj in instance_list],
            [obj.quantity for obj in instance_list],
            [obj.sky_area for obj in instance_list],
        ]

        table = Table(data=cols, names=names)
        table.meta["SACCTYPE"] = "tracer"
        table.meta["SACCCLSS"] = cls.tracer_type
        table.meta["EXTNAME"] = f"tracer:{cls.tracer_type}"
        return [table]

    @classmethod
    def from_tables(cls, table_list):
        """Convert an astropy table into a dictionary of tracers

        This is used when loading data from a file.
        One tracer object is created for each "row" in each table.

        :param table_list: List of astropy tables
        :return: Dictionary of tracers
        """
        tracers = {}

        for table in table_list:
            for row in table:
                name = row["name"]
                quantity = row["quantity"]
                sky_area = row["sky_area"]
                tracers[name] = cls(
                    name,
                    quantity=quantity,
                    sky_area=sky_area,
                )
        return tracers



from astropy.io import fits
from astropy.table import Table
import scipy.linalg
import numpy as np

from .utils import invert_spd_matrix


class BaseCovariance:
    """
    The abstract base class for covariances in different forms.
    These are not currently designed to be modified after creation.

    The three concrete subclasses that are created are:

    FullCovariance - for dense matrices

    BlockDiagonalCovariance - for block diagonal matrices
        (those in which some sub-blocks are dense but without correlation
        between the blocks

    DiagonalCovariance - a covariance where the elements are uncorrelated

    Attributes
    ----------
    cov_type: string
        The type of the covariance (class variable)
    """
    _covariance_classes = {}

    def __init__(self):
        """Abstract superclass constructor.

        All the subclasses need _dense and _dense_inverse forms.

        """
        self._dense = None
        self._dense_inverse = None

    # This method gets called whenever a subclass is
    # defined.  The keyword argument in the class definition
    # (e.g. cov_type='full' below is passed to this class method)
    @classmethod
    def __init_subclass__(cls, cov_type):
        cls._covariance_classes[cov_type] = cls
        cls.cov_type = cov_type

    @classmethod
    def from_hdu(cls, hdu):
        """
        Make a covariance object from an astropy FITS HDU object.

        The type of the covariance is determined from a keyword
        in the HDU, and then the corresponding subclass from_hdu
        method is called.

        Parameters
        ----------
        hdu: astropy.fits.ImageHDU instance
            An HDU object with covariance info in it

        Returns
        -------
        instance: BaseCovariance
            A covariance instance
        """
        subclass_name = hdu.header['saccclss']
        subclass = cls._covariance_classes[subclass_name]
        return subclass.from_hdu(hdu)

    @classmethod
    def make(cls, cov):
        """Make an appropriate covariance object from the matrix info itself.

        You can pass in a list of covariance blocks for a block-diagonal,
        covariance a 1D array for a diagonal covariance, or a full matrix.

        A different subclass is returned for each of these cases.

        Parameters
        ----------
        cov: list[array] or array
            If a list, the total length of all the arrays in it
            should equal n.  If an array, it should be either 1D of
            length n or 2D of shape (n x n).

        n: int
            length of the data vector to which this covariance applies
        """
        if isinstance(cov, list):
            s = 0
            for block in cov:
                block = np.atleast_2d(block)
                if (block.ndim != 2) or (block.shape[0] != block.shape[1]):
                    raise ValueError("Covariance block has wrong size "
                                     f"or shape {block.shape}")
                s += block.shape[0]
            return BlockDiagonalCovariance(cov)
        else:
            cov = np.array(cov).squeeze()
            if cov.ndim == 0:
                return DiagonalCovariance(np.atleast_1d(cov))
            if cov.ndim == 1:
                return DiagonalCovariance(cov)
            if (cov.ndim != 2) or (cov.shape[0] != cov.shape[1]):
                raise ValueError("Covariance is not a 2D square matrix "
                                 f"- shape: {cov.shape}")
            return FullCovariance(cov)

    @property
    def dense(self):
        """
        A dense matrix form of the covariance

        Parameters
        ----------
        None

        Returns
        -------
        covmat: 2D array
            Numpy array of dense form of matrix
        """
        if self._dense is None:
            self._dense = self._get_dense()

        return self._dense

    @property
    def inverse(self):
        """A dense matrix form of the inverse of the covariance matrix

        Returns
        -------
        invC: array
            Inverse covariance
        """
        if self._dense_inverse is None:
            self._dense_inverse = self._get_dense_inverse()
        return self._dense_inverse


class FullCovariance(BaseCovariance, cov_type='full'):
    """
    A covariance subclass representing a full matrix with correlations
    anywhere.  Represented as an n x n matrix.

    Attributes
    ----------
    size: int
        the length of the corresponding data vector

    covmat: 2D array
        The matrix itself, of shape (size x size)
    """
    def __init__(self, covmat):
        self.covmat = np.atleast_2d(covmat)
        self.size = self.covmat.shape[0]
        super().__init__()

    def to_hdu(self):
        """
        Make an astropy FITS HDU object with this covariance in it.
        This is represented as an image.

        Parameters
        ----------
        None

        Returns
        -------
        hdu: astropy.fits.ImageHDU instance
            HDU that can be used to reconstruct the object.
        """
        hdu = fits.ImageHDU(self.covmat)
        hdu.header['EXTNAME'] = 'covariance'
        hdu.header['SACCTYPE'] = 'cov'
        hdu.header['SACCCLSS'] = self.cov_type
        hdu.header['SIZE'] = self.size
        return hdu

    @classmethod
    def from_hdu(cls, hdu):
        """
        Load a covariance object from the data in the HDU

        Parameters
        ----------
        hdu: astropy.fits.ImageHDU instance

        Returns
        -------
        cov: FullCovariance
            Loaded covariance object
        """
        C = hdu.data
        return cls(C)

    def keeping_indices(self, indices):
        """
        Return a new instance with only the specified indices retained.

        Parameters
        ----------
        indices: array or list
            Either an array or list of integer indices, or a boolean
            array of the same size (1D) as the matrix.
            Specifies rows/cols to keep in the new matrix.

        Returns
        -------
        cov: FullCovariance
            A covariance with only the corresponding data points remaining
        """
        C = self.covmat[indices][:, indices]
        return self.__class__(C)

    def get_block(self, indices):
        """Read a (not necessarily contiguous) sublock of the matrix

        Parameters
        ----------
        indices: array
            An array of integer indices

        Returns
        -------
        block: array
            a 2D array of the relevant sub-block of the matrix
        """
        return self.covmat[indices][:, indices]

    def _get_dense_inverse(self):
        return invert_spd_matrix(self.covmat)

    def _get_dense(self):
        # Internal method to get a dense form of the matrix.
        # Use the property Covariance.dense instead of calling this
        # directly.
        return self.covmat.copy()


class BlockDiagonalCovariance(BaseCovariance, cov_type='block'):
    """A covariance subclass representing block diagonal covariances

    Block diagonal covariances have sub-blocks that are full dense matrices,
    but without correlations between the blocks. This feature can be taken
    advantage of when doing matrix operations like multiplication or inversion.

    Parameters
    ----------
    blocks: list[arrays]
        list of sub-blocks of the matrix

    block_sizes: list[int]
        list of sizes n of each the n x n sub-blocks

    size: int
        overall total size of the matrix
    """
    def __init__(self, blocks):
        """Create a BlockDiagonalCovariance object from a list of blocks

        Parameters
        ----------
        blocks: sequence of arrays
            List or other sequence of the sub-matrices
        """
        self.blocks = [np.atleast_2d(B) for B in blocks]
        self.block_sizes = [len(B) for B in self.blocks]
        self.size = sum(self.block_sizes)
        super().__init__()

    def to_hdu(self):
        """Write a FITS HDU from the data, ready to be saved.

        The data in the HDU is stored as a single 1 x size image,
        and the header contains the information needed to reconstruct it.

        Parameters
        ----------
        None

        Returns
        -------
        hdu: astropy.fits.ImageHDU object
            HDU containing data and metadata
        """
        hdu = fits.ImageHDU(np.concatenate([b.flatten() for b in self.blocks]))
        hdu.name = 'covariance'
        hdu.header['sacctype'] = 'cov'
        hdu.header['saccclss'] = self.cov_type
        hdu.header['size'] = self.size
        hdu.header['blocks'] = len(self.blocks)
        for i, s in enumerate(self.block_sizes):
            hdu.header[f'size_{i}'] = s
        return hdu

    @classmethod
    def from_hdu(cls, hdu):
        """Read a covariance object from a loaded FITS HDU.

        Parameters
        ----------
        hdu: FITS HDU object as read in by astropy.

        Returns
        -------
        cov: BlockDiagonalCovariance
            Loaded covariance object
        """
        n = hdu.header['blocks']
        block_sizes = [hdu.header[f'size_{i}'] for i in range(n)]
        s = 0

        blocks = []
        for b in block_sizes:
            B = hdu.data[s:s + b**2].reshape((b, b))
            s += b**2
            blocks.append(B)
        return cls(blocks)

    def get_block(self, indices):
        """Read a (not necessarily contiguous) sublock of the matrix

        Parameters
        ----------
        indices: array
            An array of integer indices, which must be in
            ascending order

        Returns
        -------
        cov: array
            A full (dense) 2x2 array of the submatrix.
        """
        indices = np.array(indices)

        if np.any(np.diff(indices)) < 0:
            raise ValueError("Indices passed to "
                             "BlockDiagonalCovariance.get_block "
                             "must be in ascending order")
        s = 0
        sub_blocks = []
        for block, sz in zip(self.blocks, self.block_sizes):
            e = s + sz
            m = indices[(indices >= s) & (indices < e)] - s
            sub_blocks.append(block[m][:, m])
            s += sz
        return scipy.linalg.block_diag(*sub_blocks)

    def keeping_indices(self, indices):
        """
        Return a new instance with only the specified elements retained.

        This method will try to return another BlockDiagonalCovariance if
        it can, but otherwise will revert to a full one: if the mask passed
        in is of a boolean type or if it is integers in it can remain
        block diagonal

        Parameters
        ----------
        indices: array or list
            Either an array or list of integer indices, or a boolean
            array of the same size (1D) as the matrix.
            Specifies rows/cols to keep in the new matrix.

        Returns
        -------
        cov: FullCovariance or BlockDiagonalCovariance
            A covariance with only the corresponding data points remaining
        """
        indices = np.array(indices)

        if indices.dtype == bool:
            breaks = np.cumsum(self.block_sizes)[:-1]
            block_masks = np.split(indices, breaks)
            blocks = [self.blocks[i][m][:, m] for i, m in
                      enumerate(block_masks)]
            return self.__class__(blocks)
        elif (np.diff(indices) > 0).all():
            s = 0
            sub_blocks = []
            for block, sz in zip(self.blocks, self.block_sizes):
                e = s + sz
                m = indices[(indices >= s) & (indices < e)] - s
                sub_blocks.append(block[m][:, m])
                s += sz
            return self.__class__(sub_blocks)
        else:
            C = scipy.linalg.block_diag(*self.blocks)
            C = C[indices][:, indices]
            return FullCovariance(C)

    def _get_dense_inverse(self):
        # Invert all the blocks individually and then
        # connect them all together
        return scipy.linalg.block_diag(*[invert_spd_matrix(B)
                                         for B in self.blocks])

    def _get_dense(self):
        # Internal method to get a dense form of the matrix.
        # Use the property Covariance.dense instead of calling this
        # directly.
        return scipy.linalg.block_diag(*self.blocks)


class DiagonalCovariance(BaseCovariance, cov_type='diagonal'):
    """A covariance subclass representing covariances that are
    purely diagonal.

    Parameters
    ----------
    size: int
        The size of the matrix

    diag: array
        The diagonal terms in the covariance (i.e. the variances)
    """
    def __init__(self, variances):
        """
        Create a DiagonalCovariance object from the variances
        of the data points.

        Parameters
        ----------
        variances: array
            2D array of variances of the data points.
        """
        self.diag = np.atleast_1d(variances)
        self.size = len(self.diag)
        super().__init__()

    def to_hdu(self):
        """
        Make an astropy FITS HDU object with this covariance in it.
        In this can a binary table HDU is created.

        Parameters
        ----------
        None

        Returns
        -------
        hdu: astropy.fits.BinTableHDU instance
            HDU that can be used to reconstruct the object.
        """
        table = Table(names=['variance'], data=[self.diag])
        hdu = fits.table_to_hdu(table)
        hdu.name = 'covariance'
        hdu.header['sacctype'] = 'cov'
        hdu.header['saccclss'] = self.cov_type
        return hdu

    def keeping_indices(self, indices):
        """
        Return a new DiagonalCovariance with only the specified indices
        retained.

        Parameters
        ----------
        indices: array or list
            Either an array or list of integer indices, or a boolean
            array of the same size (1D) as the matrix.
            Specifies rows/cols to keep in the new matrix.

        Returns
        -------
        cov: DiagonalCovariance
            A covariance with only the corresponding data points remaining
        """
        D = self.diag[indices]
        return self.__class__(D)

    @classmethod
    def from_hdu(cls, hdu):
        """
        Load a covariance object from the data in the HDU

        Parameters
        ----------
        hdu: astropy.fits.BinTableHDU instance

        Returns
        -------
        cov: DiagonalCovariance
            Loaded covariance object
        """
        D = hdu.data['variance']
        return cls(D)

    def get_block(self, indices):
        """Read a (not necessarily contiguous) sublock of the matrix

        Parameters
        ----------
        indices: array
            An array of integer indices, which should be in
            ascending order (for consistency with the
            block diagonal interface)

        Returns
        -------
        cov: array
            A full (dense) 2x2 array of the submatrix.
        """
        return np.diag(self.diag[indices])

    def _get_dense_inverse(self):
        # Trivial inverse
        return np.diag(1.0/self.diag)

    def _get_dense(self):
        # Internal method to get a dense form of the matrix.
        # Use the property Covariance.dense instead of calling this
        # directly.
        return np.diag(self.diag)


def concatenate_covariances(*covariances):
    # If all the covariances are diagonal then the concatenated
    # version can be diagonal
    if all(isinstance(cov, DiagonalCovariance) for cov in covariances):
        variances = np.concatenate([cov.diag for cov in covariances])
        return DiagonalCovariance(variances)

    # Otherwise we have to get things in a common form, and
    # make a block-diagonal covariance.
    blocks = []

    # For each of the pieces we extract any blocks
    # that will go into the concatenation
    for cov in covariances:
        # For an existing block-diagonal covariance
        # we retain the block structure
        if isinstance(cov, BlockDiagonalCovariance):
            blocks += cov.blocks
        # For everything else we just use a dense matrix
        else:
            blocks.append(cov.dense)

    return BlockDiagonalCovariance(blocks)



import numpy as np
from astropy.table import Table


class BaseWindow:
    """Base class for window functions.

    Window functions here are for 1D variables and describe
    binnings from functions onto discrete values.

    They are treated as a special kind of tag value in Sacc.

    Subclasses describe common types of window function, including
    a top hat between two values and a general tabulated
    functional form.

    This base class has class methods that can be used to turn
    mixed lists of windows to/from astropy tables, for I/O.
    """
    _window_classes = {}

    def __init_subclass__(cls, window_type):
        # This gets called whenever a subclass is defined.
        # The window_type argument is specified next to the
        # base class in the subclass definition, e.g.
        # window_typ='TopHat', as shown below
        cls._window_classes[window_type] = cls
        cls.window_type = window_type

    @classmethod
    def to_tables(cls, instance_list):
        """Convert a list of BaseWindos to a list of tables.

        This is called when saving data to file.

        The input instances can be different subclasses, and no
        ordering is maintained.

        Parameters
        ----------
        instance_list: list
            List of BaseWindow subclass instances

        Returns
        -------
        table: list
            List of astropy.table.Table instances
        """
        tables = []
        for name, subcls in cls._window_classes.items():
            # Pull out the relevant objects for this subclass.
            # Note that we can't use isinstance here.
            windows = [w for w in instance_list if type(w) == subcls]
            tables += subcls.to_tables(windows)
        return tables

    @classmethod
    def from_tables(cls, table_list):
        """Turn a list of astropy tables into window objects

        This is called when loading data from file.

        Parameters
        ----------
        instance_list: list
            List of BaseWindow instances

        Returns
        -------
        windows: dict
            Dictionary of id -> Window instances
        """
        windows = {}
        for table in table_list:
            subclass_name = table.meta['SACCCLSS']
            subclass = cls._window_classes[subclass_name]
            # Different subclasses can handle this differently.
            windows.update(subclass.from_table(table))
        return windows


class TopHatWindow(BaseWindow, window_type='TopHat'):
    """A window function that is constant between two values.

    The top-hat is zero elsewhere.

    In the case of discrete functions like ell window where it matters, these
    window functions should follow the python convention and
    the upper limit should be exclusive, so that you can use
    range(win.min, win.max) to select the right ell values.

    Parameters
    ----------
    min: int/float
        The minimum value where the top-hat function equals 1

    max: int/float
        The maximum value where the top-hat function equals 1
    """
    def __init__(self, range_min, range_max):
        """Create a top-hat window

        Parameters
        ----------
        range_min: int/float
            The minimum value where the top-hat function equals 1

        range_max: int/float
            The maximum value where the top-hat function equals 1

        """
        self.min = range_min
        self.max = range_max

    @classmethod
    def to_tables(cls, window_list):
        """Convert a list of Top-Hat windows to a list of astropy tables.

        A single table is created for all the windows.

        The tables contain an ID column which uniquely identifies the
        window instance, but only whilst running a given python process -
        it is not portable.  It should not be used for anythng other than I/O.

        Parameters
        ----------
        instance_list: list
            List of TopHatWindow instances

        Returns
        -------
        table: list
            List of astropy.table.Table instances
        """
        mins = [w.min for w in window_list]
        maxs = [w.max for w in window_list]
        ids = [id(w) for w in window_list]
        t = Table(data=[ids, mins, maxs], names=['id', 'min', 'max'])
        t.meta['SACCTYPE'] = 'window'
        t.meta['SACCCLSS'] = cls.window_type
        t.meta['EXTNAME'] = 'window:' + cls.window_type
        return [t]

    @classmethod
    def from_table(cls, table):
        """Convert an astropy table to a dictionary of Top-Hat windows

        Multiple windows are extracted from the single table.

        Parameters
        ----------
        table: astropy.table.Table

        Returns
        -------
        windows: dict
            Dictionary of id -> TopHatWindow instances
        """
        return {row['id']: cls(row['min'], row['max']) for row in table}


class LogTopHatWindow(TopHatWindow, window_type='LogTopHat'):
    """A window function that is log-constant between two values.

    This object is the same as the TopHat form, except that in between
    the min and max values it is assumed to be constant in the log of the
    argument.  The difference arises when this object is used elsewhere.
    """
    pass


class Window(BaseWindow, window_type='Standard'):
    """The Window class defines a tabulated window function.

    The class contains tabulated values of the abscissa (e.g. ell or theta) and
    corresponding weights values for each one.

    The function could be integrated or summed depending on the
    context.

    Parameters
    ----------
    values: array
        The points at which the weights are defines
    weight:
        The weights corresponding to each value

    """
    def __init__(self, values, weight):
        self.values = np.array(values)
        self.weight = np.array(weight)

    @classmethod
    def to_tables(cls, window_list):
        """Convert a list of windows to a list of astropy tables.

        One table is created per window.

        The tables contain an ID column which uniquely identifies the
        window instance, but only whilst running a given python process -
        it is not portable.  It should not be used for anythng other than I/O.

        Parameters
        ----------
        instance_list: list
            List of Window instances

        Returns
        -------
        table: list
            List of astropy.table.Table instances
        """
        tables = []
        for w in window_list:
            cols = [w.values, w.weight]
            names = ['values', 'weight']
            t = Table(data=cols, names=names)
            t.meta['SACCTYPE'] = 'window'
            t.meta['SACCCLSS'] = cls.window_type
            t.meta['SACCNAME'] = id(w)
            t.meta['EXTNAME'] = 'window:' + cls.window_type
            tables.append(t)
        return tables

    @classmethod
    def from_table(cls, table):
        """Convert an astropy table to a dictionary of Windows

        One window is extracted from the table.

        Parameters
        ----------
        table: astropy.table.Table

        Returns
        -------
        windows: dict
            Dictionary of id -> Window instances
        """
        return {table.meta['SACCNAME']: cls(table['values'], table['weight'])}


class BandpowerWindow(BaseWindow, window_type='Bandpower'):
    """The BandpowerWindow class defines a tabulated for power
    spectrum bandpowers.

    The class contains tabulated values of ell and corresponding weights
    values for each one. More than one set of weights can be used.

    Parameters
    ----------
    values: array
        An array of dimension (N_ell) containing the ell values at which the
        weights are defined.
    weight:
        An array of dimensions (N_ell, N_weight) containing N_weight sets of
        weights corresponding to each value.

    """
    def __init__(self, values, weight):
        nl, nv = weight.shape
        nell = len(values)
        if nl != len(values):
            raise ValueError(f"Wrong input shapes ${nl}!=${nell}")
        self.nell = nell
        self.nv = nv
        self.values = np.array(values)
        self.weight = np.array(weight)

    @classmethod
    def to_tables(cls, window_list):
        """Convert a list of windows to a list of astropy tables.

        One table is created per window.

        The tables contain an ID column which uniquely identifies the
        window instance, but only whilst running a given python process -
        it is not portable.  It should not be used for anythng other than I/O.

        Parameters
        ----------
        instance_list: list
            List of Window instances

        Returns
        -------
        table: list
            List of astropy.table.Table instances
        """
        tables = []
        for w in window_list:
            cols = [w.values, w.weight]
            names = ['values', 'weight']
            t = Table(data=cols, names=names)
            t.meta['SACCTYPE'] = 'window'
            t.meta['SACCCLSS'] = cls.window_type
            t.meta['SACCNAME'] = id(w)
            t.meta['EXTNAME'] = 'window:' + cls.window_type
            tables.append(t)
        return tables

    @classmethod
    def from_table(cls, table):
        """Convert an astropy table to a dictionary of Windows

        One window is extracted from the table.

        Parameters
        ----------
        table: astropy.table.Table

        Returns
        -------
        windows: dict
            Dictionary of id -> Window instances
        """
        return {table.meta['SACCNAME']: cls(table['values'], table['weight'])}

    def get_section(self, indices):
        """Get part of this window function corresponding to the input
        indices.

        Parameters
        ----------
        indices: int or array_like
            Indices to return.

        Returns
        -------
        window: `Window`
            A `Window object.
        """
        return self.__class__(self.values, self.weight[:, indices])



import warnings
from collections import namedtuple
from astropy.table import Table

from .utils import (Namespace, hide_null_values,
                    null_values, camel_case_split_and_lowercase)

# The format for a data type name looks like this:
# {sources}_{properties}_{statistic_type}[_{statistic_subtype}]
#     sources: type(s) of astrophysical sources to which this applies
#     properties: feature(s)/characterisic(s) of those sources/fields to
#         which the statistic applies
#     statistic_type: mathematical type of the statistic
#     statistic_subtype: optional additional specifier

required_tags_concise = {
    "cl_00": ['ell'],
    "cl_0e": ['ell'],
    "cl_0b": ['ell'],
    "cl_e0": ['ell'],
    "cl_b0": ['ell'],
    "cl_ee": ['ell'],
    "cl_eb": ['ell'],
    "cl_be": ['ell'],
    "cl_bb": ['ell'],
    "xi_00": ['theta'],
    "xi_0e": ['theta'],
    "xi_0b": ['theta'],
    "xi_e0": ['theta'],
    "xi_b0": ['theta'],
    "xi_plus_re": ['theta'],
    "xi_plus_im": ['theta'],
    "xi_minus_re": ['theta'],
    "xi_minus_im": ['theta'],
    "count": [],
    "cluster_counts": [],
    "cluster_mean_log_mass": [],
    "cluster_shear": []
}

required_tags_verbose = {
    "clusterGalaxy_densityConvergence_cl": ['ell'],
    "clusterGalaxy_densityConvergence_xi": ['theta'],
    "clusterGalaxy_densityShear_cl_b": ['ell'],
    "clusterGalaxy_densityShear_cl_e": ['ell'],
    "clusterGalaxy_densityShear_xi_t": ['theta'],
    "clusterGalaxy_densityShear_xi_x": ['theta'],
    "clusterGalaxy_density_cl": ['ell'],
    "clusterGalaxy_density_xi": ['theta'],
    "cluster_density_cl": ['ell'],
    "cluster_density_xi": ['theta'],
    "cluster_mass_count_wl": [],
    "cluster_mass_count_xray": [],
    "cmbCluster_polarizationDensity_cl_b": ['ell'],
    "cmbCluster_polarizationDensity_cl_e": ['ell'],
    "cmbCluster_polarizationDensity_xi_t": ['theta'],
    "cmbCluster_polarizationDensity_xi_x": ['theta'],
    "cmbCluster_temperatureDensity_cl": ['ell'],
    "cmbCluster_temperatureDensity_xi": ['theta'],
    "cmbCluster_convergenceDensity_cl": ['ell'],
    "cmbCluster_convergenceDensity_xi": ['theta'],
    "cmbGalaxy_convergenceDensity_cl": ['ell'],
    "cmbGalaxy_convergenceDensity_xi": ['theta'],
    "cmbGalaxy_convergenceShear_cl_b": ['ell'],
    "cmbGalaxy_convergenceShear_cl_e": ['ell'],
    "cmbGalaxy_convergenceShear_xi_t": ['theta'],
    "cmbGalaxy_convergenceShear_xi_x": ['theta'],
    "cmbGalaxy_convergence_cl": ['ell'],
    "cmbGalaxy_convergence_xi": ['theta'],
    "cmbGalaxy_polarizationConvergence_cl_b": ['ell'],
    "cmbGalaxy_polarizationConvergence_cl_e": ['ell'],
    "cmbGalaxy_polarizationConvergence_xi_t": ['theta'],
    "cmbGalaxy_polarizationConvergence_xi_x": ['theta'],
    "cmbGalaxy_polarizationDensity_cl_b": ['ell'],
    "cmbGalaxy_polarizationDensity_cl_e": ['ell'],
    "cmbGalaxy_polarizationDensity_xi_t": ['theta'],
    "cmbGalaxy_polarizationDensity_xi_x": ['theta'],
    "cmbGalaxy_polarizationShear_cl_bb": ['ell'],
    "cmbGalaxy_polarizationShear_cl_be": ['ell'],
    "cmbGalaxy_polarizationShear_cl_eb": ['ell'],
    "cmbGalaxy_polarizationShear_cl_ee": ['ell'],
    "cmbGalaxy_polarizationShear_xi_minus": ['theta'],
    "cmbGalaxy_polarizationShear_xi_plus": ['theta'],
    "cmbGalaxy_polarizationShear_xi_imagMinus": ['theta'],
    "cmbGalaxy_polarizationShear_xi_imagPlus": ['theta'],
    "cmbGalaxy_polarizationShear_cosebi_e": ['ell'],
    "cmbGalaxy_polarizationShear_cosebi_b": ['ell'],
    "cmbGalaxy_temperatureConvergence_cl": ['ell'],
    "cmbGalaxy_temperatureConvergence_xi": ['theta'],
    "cmbGalaxy_temperatureDensity_cl": ['ell'],
    "cmbGalaxy_temperatureDensity_xi": ['theta'],
    "cmbGalaxy_temperatureShear_cl_b": ['ell'],
    "cmbGalaxy_temperatureShear_cl_e": ['ell'],
    "cmbGalaxy_temperatureShear_xi_t": ['theta'],
    "cmbGalaxy_temperatureShear_xi_x": ['theta'],
    "cmb_convergence_cl": ['ell'],
    "cmb_convergence_xi": ['theta'],
    "cmb_polarization_cl_bb": ['ell'],
    "cmb_polarization_cl_be": ['ell'],
    "cmb_polarization_cl_eb": ['ell'],
    "cmb_polarization_cl_ee": ['ell'],
    "cmb_polarization_xi_minus": ['theta'],
    "cmb_polarization_xi_plus": ['theta'],
    "cmb_polarization_xi_imagMinus": ['theta'],
    "cmb_polarization_xi_imagPlus": ['theta'],
    "cmb_polarization_cosebi_e": [],
    "cmb_polarization_cosebi_b": [],
    "cmb_polarizationConvergence_cl_e": ['ell'],
    "cmb_polarizationConvergence_cl_b": ['ell'],
    "cmb_polarizationConvergence_xi_t": ['theta'],
    "cmb_polarizationConvergence_xi_x": ['theta'],
    "cmb_temperaturePolarization_cl_e": ['ell'],
    "cmb_temperaturePolarization_cl_b": ['ell'],
    "cmb_temperaturePolarization_xi_t": ['theta'],
    "cmb_temperaturePolarization_xi_x": ['theta'],
    "cmb_temperature_cl": ['ell'],
    "cmb_temperatureConvergence_cl": ["ell"],
    "cmb_temperatureConvergence_xi": ["theta"],
    "galaxy_convergenceDensity_cl": ['ell'],
    "galaxy_convergenceDensity_xi": ['theta'],
    "galaxy_convergence_cl": ['ell'],
    "galaxy_convergence_xi": ['theta'],
    "galaxy_density_cl": ['ell'],
    "galaxy_density_xi": ['theta'],
    "galaxy_shearConvergence_cl_b": ['ell'],
    "galaxy_shearConvergence_cl_e": ['ell'],
    "galaxy_shearConvergence_xi_t": ['theta'],
    "galaxy_shearConvergence_xi_x": ['theta'],
    "galaxy_shearDensity_cl_b": ['ell'],
    "galaxy_shearDensity_cl_e": ['ell'],
    "galaxy_shearDensity_xi_t": ['theta'],
    "galaxy_shearDensity_xi_x": ['theta'],
    "galaxy_shear_cl_bb": ['ell'],
    "galaxy_shear_cl_be": ['ell'],
    "galaxy_shear_cl_eb": ['ell'],
    "galaxy_shear_cl_ee": ['ell'],
    "galaxy_shear_cosebi_bb": [],
    "galaxy_shear_cosebi_ee": [],
    "galaxy_shear_xi_minus": ['theta'],
    "galaxy_shear_xi_plus": ['theta'],
    "galaxy_shear_xi_imagMinus": ['theta'],
    "galaxy_shear_xi_imagPlus": ['theta'],
}

required_tags = {**required_tags_concise, **required_tags_verbose}

parsedDataTypeName = namedtuple('parsedDataTypeName',
                                'sources properties statistic subtype')


def parse_data_type_name(name):
    """Parse a verbose data type name into its component parts
    Verbose data type names take the form:
    {sources}_{properties}_{statistic_type}[_{statistic_subtype}]
    where sources and properties are camel-case if there is more
    than one of them

    Parameters
    ----------
    name: str
        A data type name

    Returns
    -------
    sources: list[str]
        type(s) of astrophysical sources to which this applies

    properties: list[str]
        feature(s)/characterisic(s) of those sources/fields to
        which the statistic applies

    statistic_type: str
        mathematical type of the statistic

    statistic_subtype: str or None
        optional additional specifier
    """
    parts = name.split("_")
    if len(parts) == 3:
        sources, properties, statistic = parts
        subtype = None
    elif len(parts) == 4:
        sources, properties, statistic, subtype = parts
    else:
        raise ValueError("The supplied name is not a valid "
                         "verbose data type name (must have "
                         f"3 or 4 underscore-sparated parts): {name}")
    sources = camel_case_split_and_lowercase(sources)
    properties = camel_case_split_and_lowercase(properties)
    return parsedDataTypeName(sources, properties, statistic, subtype)


def build_data_type_name(sources, properties, statistic, subtype=None):
    """
    Parameters
    ----------
    sources: str or list[str]
        type(s) of astrophysical sources to which this applies

    properties: str or list[str]
        feature(s)/characterisic(s) of those sources/fields to
        which the statistic applies

    statistic_type: str
        mathematical type of the statistic

    statistic_subtype: str or None
        optional additional specifier.  Default is None

    Returns
    -------
    name: str
        Type name of the form:
           {sources}_{properties}_{statistic_type}[_{statistic_subtype}]
    """
    if not isinstance(sources, str):
        sources = "".join([sources[0]] + [s.lower().capitalize()
                                          for s in sources[1:]])
    if not isinstance(properties, str):
        properties = "".join([properties[0]] + [s.lower().capitalize()
                                                for s in properties[1:]])
    if subtype:
        return f"{sources}_{properties}_{statistic}_{subtype}"
    else:
        return f"{sources}_{properties}_{statistic}"


# This makes a namespace object, so you can do:
# standard_types.ggl_e == "ggl_e"
# also, for convenience, you can do standard_types.index('ggl_e')
# and 'ggl_e' in standard_types

standard_types = Namespace(*required_tags.keys())


class DataPoint:
    """A class for a single data point (one scalar value).

    Data points have a type, zero or more tracers, a value,
    and any arbitrary tags that are stored in a dictionary,
    and can be used to describe angular scales, window functions,
    or any arbitrary information to be attached to the data.

    Data points can be automatically created and added to a
    Sacc object, so you don't normally nee to manually create them.

    Parameters
    -----------
    data_type: str
        A string, indicating the type of data point

    tracers: tuple
        Tuple of strings with the names of tracers to use

    value: float
        Mean value of this statistics

    tags: dict
        Dictionary of further data point metadata, such as binning
        info, angles, etc.
    """
    def __init__(self, data_type, tracers, value,
                 ignore_missing_tags=False, **tags):
        """Create a new data point.

        Data points can be automatically created and added to a
        Sacc object, so you don't normally nee to manually create them.

        Parameters
        ----------
        data_type: str
            A string, indicating the type of data point

        tracers: tuple
            Tuple of strings with the names of tracers to use

        value: float
            Mean value of this statistics

        ignore_missing_tags: bool
            Optional, default=False.  If True, do not complain if a
            tracer usually needed for this data type is not present.

        **tags: dict[str:any]
            Dictionary of further data point metadata, such as binning
            info, angles, etc.
        """
        self.data_type = data_type
        self.tracers = tracers
        self.value = value
        self.tags = tags
        # Data types can have required tags which must be present.
        # Check for those here
        if (data_type in required_tags) and (not ignore_missing_tags):
            for tag in required_tags[data_type]:
                if tag not in tags:
                    raise ValueError("Tag {tag} required for data type "
                                     f"{data_type} "
                                     "(ignore_missing_tags=False)")

        # We encourage people to use existing type names, and issue a
        # warning if they do not to prod them in the right direction.
        if data_type not in standard_types:
            warnings.warn(f"Unknown data_type value {data_type}. "
                          "If possible use a pre-defined type, or "
                          "add to the list.")

    def __repr__(self):
        t = ", ".join(f'{k}={v}' for (k, v) in self.tags.items())
        st = f"DataPoint(data_type='{self.data_type}', "
        st += f"tracers={self.tracers}, value={self.value}, {t})"
        return st

    def get_tag(self, tag, default=None):
        """
        Get the value of the the named tag, or None if not found.

        Parameters
        ----------
        tag: str
            Tag to find on the data point

        default: any
            Value to return if the tag is not found

        Returns
        -------
        value: any
            Value of the tag in this data point
        """
        return self.tags.get(tag, default)

    def __getitem__(self, tag):
        """
        Get the value of the the named tag, raising an
        error if it is not found

        Parameters
        ----------
        tag: str
            Tag to find on the data point

        Returns
        -------
        value: any
            Value of the tag in this data point
        """
        return self.tags[tag]

    @staticmethod
    def _choose_fields(data):
        """
        Internal static method to generate a list of colum names from a list
        of data points.  Since the data points can be heterogenous then this
        is not quite trivial - we use the union of the tag names and tracer_0,
        tracer_1, etc. up to the max number of tracers.
        """
        tags = set()
        ntracer = 0
        for d in data:
            ntracer = max(ntracer, len(d.tracers))
            tags.update(d.tags.keys())
        tags = list(tags)
        tracers = [f'tracer_{i}' for i in range(ntracer)]
        return tracers, tags

    @classmethod
    def to_table(cls, data, lookups=None):
        """
        Convert a list of data points to a single homogenous table

        Since data points can have varying tags, this method uses
        null values to represent non-present tags.

        Parameters
        ----------
        data: list
            A list of DataPoint objects

        lookups: dict
            A dictionary of tags->dict showing replacements to make
            in the tags. Default is empty.

        Returns
        -------
        table: astropy.table.Table
            table object containing data points
        """
        if lookups is None:
            lookups = {}
        # Get the names of the columns to generate
        tracers, tags = cls._choose_fields(data)
        names = tracers + ['value'] + tags
        ntracer = len(tracers)
        # Convert each data point to a row
        rows = [d._make_row(tracers, tags, lookups) for d in data]

        # Convert to a table and fiddle slightly.
        table = Table(rows=rows, names=names)
        table.meta['NTRACER'] = ntracer
        hide_null_values(table)
        return table

    @classmethod
    def from_table(cls, table, lookups=None):
        """Convert a table back into a list of data points.

        This method removes null values from the tags.

        Parameters
        ----------
        table: astropy.table.Table
            A table of data containing the tracers, values, and tags

        lookups: dict
            A dictionary of tags->dict showing replacements to make
            in the tags. Default is empty.

        Returns
        -------
        data: list
            list of DataPoint objects
        """
        if lookups is None:
            lookups = {}
        # Get out required table metadata
        nt = table.meta['NTRACER']
        data_type = table.meta['SACCNAME']

        # Tag names - we will remove missing tags below
        tag_names = table.colnames[nt + 1:]
        data = []
        for row in table:
            # Get basic data elements
            tracers = tuple([row[f'tracer_{i}'] for i in range(nt)])
            value = row['value']

            # Deal with tags.  First just pull out all remaining columns
            tags = {name: row[name] for name in tag_names}
            for k, v in list(tags.items()):
                # Deal with any tags that we should replace.
                # This is mainly used for Window instances.
                if k in lookups:
                    tags[k] = lookups[k].get(v, v)
                # Now delete and null values, as indicated by the
                # sentinel above.
                if ((hasattr(tags[k], 'dtype')) and
                        (v == null_values[tags[k].dtype.kind])):
                    del tags[k]
            # Finally convert back to a data point and record
            data_point = cls(data_type, tracers, value, **tags)
            data.append(data_point)
        return data

    def _make_row(self, tracers, tags, lookups):
        """
        Turn this data point into a list with specified tracers and tags.
        If some tracers or tags are missing (homogenous data set) then
        use blank values or Nones for them.
        """
        nt = len(tracers)
        missing = nt - len(self.tracers)
        row = list(self.tracers) + ["" for i in range(missing)]
        row.append(self.value)
        for t in tags:
            v = self.tags.get(t)
            lookup = lookups.get(t)
            if lookup is not None:
                v = lookup.get(v, v)
            row.append(v)
        return row



import sacc
import numpy as np
from packaging import version


# Dummy arrays
vecsize = 100
x = np.linspace(0, 100, vecsize)
y = 1./(x+50)**2

# Create file
s = sacc.Sacc()

# Add tracers
s.add_tracer('NZ', 'nz', x, y)
s.add_tracer('NZ', 'nz_b', x, y)

wth = sacc.TopHatWindow(0., 1.)
ww = sacc.Window(x, y)
s.add_ell_cl('galaxy_density_cl', 'nz', 'nz', x, y, window=[wth] * vecsize)
s.add_theta_xi('galaxy_density_xi', 'nz', 'nz', x, y, window=[ww] * vecsize)
nvec = 2

if version.parse(sacc.__version__) > version.parse('0.2.0'):
    wlth = sacc.LogTopHatWindow(-2., 0.)
    s.add_ell_cl('galaxy_density_cl', 'nz_b', 'nz_b', x, y,
                 window=[wlth] * vecsize)
    nvec += 1

if version.parse(sacc.__version__) > version.parse('0.4.0'):
    wbpw = sacc.BandpowerWindow(x, np.ones([vecsize, vecsize]))
    s.add_tracer('Misc', 'msc')
    s.add_tracer('Map', 'map', 0, x, y)
    s.add_tracer('NuMap', 'nap', 2, x, y, x, y)
    s.add_ell_cl('cl_00', 'nz', 'map', x, y, window=wbpw)
    s.add_theta_xi('xi_plus_re', 'msc', 'nap', x, y, window=[wth] * vecsize)
    s.add_ell_cl('cl_00', 'map', 'nz_b', x, y, window=wbpw)
    nvec += 3

s.add_covariance(np.eye(nvec * vecsize))
fname_save = 'legacy_files/dummy_v' + sacc.__version__ + '.fits'
s.save_fits(fname_save, overwrite=True)



import re

from astropy.table import Column
import numpy as np
import scipy.linalg

# These null values are used in place
# of missing values.
null_values = {
    'i': -1437530437530211245,
    'f': -1.4375304375e30,
    'U': '',
}


def hide_null_values(table):
    """Replace None values in an astropy table
    with a null value, and change the column type
    to whatever suits the remaining data points

    Parameters
    ----------
    table: astropy table
        Table to modify in-place
    """

    for name, col in list(table.columns.items()):
        if col.dtype.kind == 'O':
            good_values = [x for x in col if x is not None]
            good_kind = np.array(good_values).dtype.kind
            null = null_values[good_kind]
            good_col = np.array([null if x is None else x for x in col])
            table[name] = Column(good_col)


def remove_dict_null_values(dictionary):
    """Remove values in a dictionary that
    correspond to the null values above.

    Parameters
    ----------
    dictionary: dict
        Dict (or subclass instance or other mapping) to modify in-place
    """
    # will figure out the list of keys to remove
    deletes = []
    for k, v in dictionary.items():
        try:
            dt = np.dtype(type(v)).kind
            if v == null_values[dt]:
                deletes.append(k)
        except (TypeError, KeyError):
            continue
    for d in deletes:
        del dictionary[d]


def unique_list(seq):
    """
    Find the unique elements in a list or other sequence
    while maintaining the order. (i.e., remove any duplicated
    elements but otherwise leave it the same)

    Method from:
    https://stackoverflow.com/questions/480214/how-do-you-remove-duplicates-from-a-list-whilst-preserving-order

    Parameters
    ----------
    seq: list or sequence
        Any input object that can be iterated

    Returns
    -------
    L: list
        a new list of the unique objects
    """
    seen = set()
    seen_add = seen.add
    return [x for x in seq if not (x in seen or seen_add(x))]


class Namespace:
    """
    This helper class implements a very simple namespace object
    designed to store strings with the same name as their contents
    as a kind of simple string enum.

    N = Namespace('a', 'b', 'c')

    assert N.a=='a'

    assert N['a']=='a'

    assert N.index('b')==1
    """
    def __init__(self, *strings):
        """
        Create the object from a list of strings, which will become attributes
        """
        self._index = {}
        n = 0
        for s in strings:
            self.__dict__[s] = s
            self._index[s] = n
            n += 1

    def __contains__(self, s):
        return hasattr(self, s)

    def __getitem__(self, s):
        return getattr(self, s)

    def __str__(self):
        return "\n".join(f"- {s}" for s in self._index)

    def index(self, s):
        return self._index[s]


def invert_spd_matrix(M, strict=True):
    """
    Invert a symmetric positive definite matrix.

    SPD matrices (for example, covariance matrices) have only
    positive eigenvalues.

    Based on:
    https://stackoverflow.com/questions/40703042/more-efficient-way-to-invert-a-matrix-knowing-it-is-symmetric-and-positive-semi

    Parameters
    ----------
    M: 2d array
        Matrix to invert

    strict: bool, default=True
        If True, require that the matrix is SPD.
        If False, use a slower algorithm that will work on other matrices

    Returns
    -------
    invM: 2d array
        Inverse matrix

    """
    M = np.atleast_2d(M)

    # In strict mode we use a method which will only work for SPD matrices,
    # and raise an exception otherwise.
    if strict:
        L, _ = scipy.linalg.lapack.dpotrf(M, False, False)
        invM, info = scipy.linalg.lapack.dpotri(L)
        if info:
            raise ValueError("Matrix is not symmetric-positive-definite")
        else:
            invM = np.triu(invM) + np.triu(invM, k=1).T
    # Otherwise we use the generic (and also slower) method that will
    # work if, due to numerical issues, the matrix is not quite SPD
    else:
        invM = np.linalg.inv(M)

    return invM


def camel_case_split_and_lowercase(identifier):
    matches = re.finditer('.+?(?:(?<=[a-z])(?=[A-Z])'
                          '|(?<=[A-Z])(?=[A-Z][a-z])|$)',
                          identifier)
    return [m.group(0).lower() for m in matches]



#!/usr/bin/env python
from setuptools import setup

description = "SACC - the LSST/DESC summary statistic data format library"

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

with open('README.md') as f:
    long_description = f.read()

setup(name="sacc", 
      description=description,
      long_description=long_description,
      long_description_content_type='text/markdown',
      url="https://github.com/LSSTDESC/sacc",
      author="LSST DESC",
      author_email="joezuntz@googlemail.com",
      use_scm_version=True,
      setup_requires=["setuptools_scm<7.0", "setuptools_scm_git_archive"],
      install_requires=requirements,
      packages=['sacc'],
)



from .sacc import Sacc, DataPoint, concatenate_data_sets  # noqa
from .windows import Window, BandpowerWindow, TopHatWindow, LogTopHatWindow  # noqa
from .data_types import standard_types, parse_data_type_name, build_data_type_name  # noqa
from .tracers import BaseTracer  # noqa
from .covariance import BaseCovariance  # noqa
__version__ = '0.9' #noqa
